"""An example module computing error as a function of stiffness.

This module contains a ``main`` function that allows the relative L2 error
to be computed for each of the asymptotic approaches, where the
approach to compare against is the exact approach, when available,
and the numerical quadrature approach otherwise.
The results are plotted using ``matplotlib``.
When executed at the command line, flags are passed as keyword arguments.

Example:
    Compute the relative L2 error when using the reduced and
    asymptotic approaches of approximating the isotensional
    single-chain mechanical response of the Morse-FJC model
    as a function of the nondimensional stiffness:

    ::

        python -m ufjc.examples.error --potential 'morse'

"""

from scipy.integrate import quad


def main(**kwargs):
    r"""Main function for the module.

    This is the main function, called when executing the module from the
    command line, also available when importing the module.

    Args:
        **kwargs: Arbitrary keyword arguments.
            Passed to ``uFJC`` instantiation.

    Example:

        .. plot::

            Plot the relative L2 error for a few different potentials:

                >>> import numpy as np
                >>> from ufjc.examples import error
                >>> error.main(potential='harmonic')
                >>> error.main(potential='log-squared')
                >>> error.main(potential='morse')
                >>> error.main(potential='lennard-jones')

    Example:
        Export .csv files for external use:

            >>> from ufjc.examples import error
            >>> error.main(potential='harmonic', csv=1)

    """

    # Import external modules
    import numpy as np
    import matplotlib.pyplot as plt

    # Import internal modules
    from ufjc import uFJC

    # Enumerate nondimensional energies
    temp = uFJC(**kwargs)
    scale = temp.varepsilon/temp.kappa
    varepsilons = np.logspace(
        np.log(scale*1e1)/np.log(10),
        np.log(scale*1e4)/np.log(10),
        int(kwargs.get('num_points', 3))
    )

    # Allocate space
    kappas = np.zeros(len(varepsilons))
    L_2_rel_error_norm_asymptotic = np.zeros(len(varepsilons))
    L_2_rel_error_norm_reduced = np.zeros(len(varepsilons))

    # Loop over nondimensional energies
    for i, varepsilon in enumerate(varepsilons):

        # Create the single-chain model
        model = uFJC(varepsilon=varepsilon, **kwargs)
        kappas[i] = model.kappa

        # Upper limit of integration
        if hasattr(model, 'lambda_max'):
            upper_lim = model.lambda_max
        else:
            upper_lim = 0.3*model.varepsilon

        # Define the function to compare with
        def gamma_0(eta):
            if model.potential == 'harmonic':
                return model.gamma(eta, approach='exact')
            else:
                return model.gamma(eta, approach='quadrature')

        # Compute the relative L_2 error norm
        L_2_rel_error_norm_asymptotic[i] = np.sqrt(
            quad(
                lambda eta: (
                    model.gamma(eta, approach='asymptotic') - gamma_0(eta)
                )**2, model.minimum_float, upper_lim
            )[0]/quad(
                lambda eta: gamma_0(eta)**2, model.minimum_float, upper_lim
            )[0]
        )
        L_2_rel_error_norm_reduced[i] = np.sqrt(
            quad(
                lambda eta: (
                    model.gamma(eta, approach='reduced') - gamma_0(eta)
                )**2, model.minimum_float, upper_lim
            )[0]/quad(
                lambda eta: gamma_0(eta)**2, model.minimum_float, upper_lim
            )[0]
        )

    # Check if opted to export .csv files
    if (kwargs.get('csv', 0) == 1) is True:
        np.savetxt(
            './fig-error-' + str(model.potential) + '.csv',
            np.vstack((
                L_2_rel_error_norm_asymptotic,
                L_2_rel_error_norm_reduced,
                kappas
            )).T,
            delimiter=","
        )

    # Plot the L_2 error norm if not opted to export .csv files
    else:
        plt.figure()
        plt.loglog(kappas, L_2_rel_error_norm_asymptotic,
                   'o-', label='asymptotic', linewidth=1.5)
        plt.loglog(kappas, L_2_rel_error_norm_reduced,
                   'o-', label='reduced', linewidth=1.5)
        plt.legend()
        plt.xlabel(r'$\kappa$')
        plt.ylabel(r'$||\gamma - \gamma_0||_2/||\gamma_0||_2$')
        plt.title(model.potential + '-FJC isotensional relative error')
        plt.show()


if __name__ == '__main__':  # pragma: no cover

    # Import external modules
    import argparse

    # Parse args
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--potential', type=str, default='harmonic')

    # Parse arbitrary float args
    for arg in parser.parse_known_args()[1]:
        if arg.startswith('--'):
            parser.add_argument(arg.split('=')[0], type=float)

    # Execute main function with passed args as kwargs
    main(**vars(parser.parse_args()))



"""Module-level functionality for running examples.

Example:
    Run the examples in the package:

    ::

        python -m ufjc.examples

    Alternatively:

        >>> from ufjc.examples import __main__

"""

from ufjc.examples import error
from ufjc.examples import approaches
from ufjc.examples import asymptotics
error.main()
approaches.main()
asymptotics.main(varepsilon_list=[10, 25, 100, 1000])



"""A module for basic single-chain model utilities.

This module consist of the class ``BasicUtility`` which contains
attributes and methods that are meant to be inherited and utilized as
basic utilities by an arbitrary single-chain model class.

Example:
    Create a freely-jointed chain single-chain model class ``FJC`` that
    inherits the ``BasicUtility`` class in order to utilize the
    Langevin function for the nondimensional isotensional mechanical response:

        >>> from ufjc.utility import BasicUtility
        >>> class FJC(BasicUtility):
        ...     def __init__(self):
        ...        BasicUtility.__init__(self)
        ...     def gamma(self, eta):
        ...         return self.langevin(eta)
        >>> model = FJC()
        >>> model.gamma([0, 0.23, 1e88])
        array([0.        , 0.07639764, 1.        ])

"""

# Import external modules
import sys
import numpy as np
from scipy.optimize import root_scalar


class BasicUtility(object):
    """The single-chain model basic utilities class.

    This class contains
    attributes and methods that are meant to be inherited and utilized as
    basic utilities by an arbitrary single-chain model class.

    Attributes:
        minimum float:
            Helps avoid overflow when dividing by small numbers.
        maximum_float:
            Helps avoid overflow in function evaluations.
        maximum_exponent:
            Helps avoid overflow in exponential functions.

    """
    def __init__(self):
        """Initializes the ``BasicUtility`` class.

        Inherits overflow-related parameters from ``sys.float_info``.
        """
        self.minimum_float = sys.float_info.min
        self.maximum_float = sys.float_info.max
        self.maximum_exponent = np.log(self.maximum_float)

    def np_array(self, input):
        """Function to return input as a numpy array.

        This function essentially serves as a wrapper for ``numpy.array``
        that returns non-array type inputs as shape (1,) numpy arrays
        rather than shape () numpy arrays, which is useful for indexing.

        Args:
            input (array_like): Anything passable to ``numpy.array``.

        Returns:
            numpy.ndarray: The input as an index-able numpy array.

        Example:
            Compare attempts to index the numpy array of an integer:

                >>> import numpy as np
                >>> from ufjc.utility import BasicUtility
                >>> try:
                ...     np.array(8)[0]
                ... except IndexError:
                ...     print('Error indexing 0-dimensional array.')
                ... finally:
                ...     BasicUtility().np_array(8)[0]
                Error indexing 0-dimensional array.
                8

        """
        if np.array(input).shape == ():
            return np.array([input])
        else:
            return np.array(input)

    def inv_fun_1D(self, y, f, power=1, guess=None, **kwargs):
        """Function to invert a mathematical function.

        This function returns the argument x given a function f(x)
        and y, the query value of the function, i.e. y = f(x).

        Note:
            This function is only meant for bijective f(x).

        Args:
            y (array_like): The value(s) of the function y.
            f (function): The function f(x).
            guess (list, optional): A list of two initial guesses.
            **kwargs: Arbitrary keyword arguments.
                Passed to ``root_scalar``.

        Returns:
            numpy.ndarray: The corresponding argument(s) x.

        Example:
            Invert a polynomial function:

                >>> from ufjc.utility import BasicUtility
                >>> f = lambda x: 1 + x**2/8 + x**3/23
                >>> BasicUtility().inv_fun_1D([55, 88], f)
                array([ 9.8712079, 11.7121826])

        """
        y = self.np_array(y)
        x = np.zeros(y.shape)
        for i in range(len(y)):

            # Function whose root will be found
            def fun(x):
                return (f(x) - y[i])**power

            # First two guesses when root-finding
            if guess is not None:
                guess_0 = guess*0.95
                guess_1 = guess
            elif y[i] == 0:
                guess_0 = 0
                guess_1 = 1e-2
            else:
                guess_0 = y[i]*0.95
                guess_1 = y[i]

            # Find the root
            x[i] = root_scalar(fun, x0=guess_0, x1=guess_1, **kwargs).root

        # Return the roots
        return x

    def log_over_sinh(self, eta):
        r"""The natural logarithm of the argument over the hyperbolic sine.

        This function avoids overflow when calculating the natural logarithm
        of the ratio of the argument to the hyperbolic sine of the argument,

        .. math::
            f(\eta) = \ln\left[\frac{\eta}{\sinh(\eta)}\right].

        It also returns zero for zero argument.

        Args:
            eta (array_like): The argument(s) of the function.

        Returns:
            numpy.ndarray: The value(s) of the function.

        Example:
            Compute the function for several argument values:

                >>> from ufjc.utility import BasicUtility
                >>> BasicUtility().log_over_sinh([0, 0.23, 888])
                array([ 0.00000000e+00, -8.80117196e-03, -8.80517881e+02])

        """
        # Determine when argument is sufficiently large
        eta = self.np_array(eta)
        eta_is_big = np.nan_to_num(eta) > 3e1
        y = np.zeros(eta.size)

        # Use asymptotic relation valid for sufficiently large arguments
        if eta_is_big.any():
            y[eta_is_big] = np.log(2*eta[eta_is_big]) - eta[eta_is_big]

        # Compute analytically otherwise, and zero where argument is zero
        where_eta_zero = eta == 0
        valid_eta = ~(eta_is_big + where_eta_zero)
        if valid_eta.any():
            y[valid_eta] = np.log(eta[valid_eta]/np.sinh(eta[valid_eta]))

        # Return the results
        return y

    def coth(self, eta):
        """The hyperbolic cotangent function.

        This function essentially serves as a wrapper for the algebraic inverse
        of ``numpy.tanh`` that avoids the singularity from evaluating at zero.

        Args:
            eta (array_like): The argument(s) of the
                hyperbolic cotangent function.

        Returns:
            numpy.ndarray: The hyperbolic cotangent(s).

        Example:
            Compute the hyperbolic cotangent for several argument values:

                >>> from ufjc.utility import BasicUtility
                >>> BasicUtility().coth([0, 0.23, np.inf])
                array([4.49423284e+307, 4.42422373e+000, 1.00000000e+000])

        """
        eta = self.np_array(eta)
        eta = np.where(eta == 0, self.minimum_float, eta)
        return 1/np.tanh(eta)

    def langevin(self, eta):
        r"""The Langevin function.

        This function computes and returns the Langevin function
        of the argument eta, which is

        .. math::
            \mathcal{L}(\eta) = \coth(\eta) - 1/\eta.

        It also avoids dividing by zero at zero argument.

        Args:
            eta (array_like): The argument(s) of the Langevin function.

        Returns:
            numpy.ndarray: The evaluated Langevin function(s).

        Example:
            Compute the Langevin function for several argument values:

                >>> from ufjc.utility import BasicUtility
                >>> BasicUtility().langevin([0, 1, 8, np.inf])
                array([0.        , 0.31303529, 0.87500023, 1.        ])

        """
        eta = self.np_array(eta)
        eta = np.where(eta == 0, self.minimum_float, eta)
        return 1/np.tanh(eta) - 1/eta

    def inv_langevin(self, gamma):
        r"""The inverse Langevin function.

        This function computes the inverse of the Langevin function
        given its query value.

        Args:
            gamma (array_like): The value(s) of the Langevin function.

        Returns:
            numpy.ndarray: The argument(s) of the Langevin function.

        Example:
            Check that :math:`\mathcal{L}^{-1}[\mathcal{L}(\eta)] = \eta\,`:

                >>> import numpy as np
                >>> from ufjc.utility import BasicUtility
                >>> def check_langevin(eta):
                ...     langevin = BasicUtility().langevin
                ...     inv_langevin = BasicUtility().inv_langevin
                ...     return np.isclose(inv_langevin(langevin(eta))[0], eta)
                >>> check_langevin(np.random.rand())
                True

        """
        return self.inv_fun_1D(gamma, self.langevin)



"""A module for the uFJC single-chain model in the isotensional ensemble.

This module consist of the class ``uFJCIsotensional`` which contains
methods for computing single-chain quantities
in the isotensional (constant end-to-end force) thermodynamic ensemble.

Example:
    Import and instantiate the class:

        >>> from ufjc.isotensional import uFJCIsotensional
        >>> class_instance = uFJCIsotensional()

"""

# Import internal modules
from .monte_carlo import MHMCMC
from .utility import BasicUtility

# Import external modules
import numpy as np
import numpy.linalg as la
from scipy.special import erf
from scipy.integrate import quad_vec


class uFJCIsotensional(BasicUtility):
    """The uFJC single-chain model class for the isotensional ensemble.

    This class contains methods for computing single-chain quantities
    in the isotensional (constant end-to-end force) thermodynamic ensemble.
    It inherits all attributes and methods from the ``BasicUtility`` class.

    """
    def __init__(self):
        """Initializes the ``uFJCIsotensional`` class.

        Initialize and inherit all attributes and methods
        from a ``BasicUtility`` class instance.

        """
        BasicUtility.__init__(self)

    def gamma_isotensional(self, eta, **kwargs):
        r"""Main function for the isotensional :math:`\gamma(\eta)`.

        This is the main function utilized to compute the isotensional
        nondimensional single-chain mechanical response.
        Keyword arguments specify and are passed onto the approach.

        Args:
            eta (array_like): The nondimensional force(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to ``gamma_isotensional_MHMCMC`` when the
                Monte Carlo approach is chosen.

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Example:
            Compute the nondimensional end-to-end length of the EFJC
            in the isotensional ensemble under a nondimensional force of 23
            using numerical quadrature:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='harmonic')
                >>> model.gamma_isotensional(23, approach='quadrature')
                array([1.22689438])

        Note:
            An improperly-specified approach will result in nan:

                >>> from ufjc import uFJC
                >>> uFJC().gamma_isotensional(0.8, approach='blah blah')
                array([nan])

        """
        eta = self.np_array(eta)
        approach = kwargs.get('approach', 'asymptotic')
        if approach == 'asymptotic':
            return self.gamma_isotensional_asymptotic(eta)
        elif approach == 'reduced':
            return self.gamma_isotensional_asymptotic_reduced(eta)
        elif approach == 'exact':
            return self.gamma_isotensional_exact(eta)
        elif approach == 'quadrature':
            return self.gamma_isotensional_quadrature(eta)
        elif approach == 'monte carlo':
            return self.gamma_isotensional_MHMCMC(eta, **kwargs)
        else:
            return np.nan*eta

    def gamma_isotensional_asymptotic(self, eta):
        r"""The full asymptotic approach for the
        isotensional :math:`\gamma(\eta)`.

        This function provides the asymptotic approach for
        obtaining :math:`\gamma(\eta)` in the isotensional ensemble
        :cite:`buche2022freely`,
        a composite approximation for all :math:`\eta`
        that is valid for :math:`\varepsilon\gg 1`,

        .. math::
            \gamma(\eta) \sim \mathcal{L}(\eta) +
                \frac{\eta}{\kappa}\left[
                \frac{1 - \mathcal{L}(\eta)\coth(\eta)}
                {c + (\eta/\kappa)\coth(\eta)}\right]
                + \Delta\lambda(\eta)
            .

        Args:
            eta (array_like): The nondimensional force(s).

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Example:
            Compute the asymptotically-correct nondimensional end-to-end length
            of the Mie-FJC in the isotensional ensemble
            under a nondimensional force of 23:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='mie')
                >>> model.gamma(23, approach='asymptotic')
                array([0.96204056])

        """
        Ln = self.langevin(eta)
        coth = self.coth(eta)
        return Ln + self.delta_lambda(eta) + \
            eta/self.kappa*(
                (1 - Ln*coth)/(self.c + eta/self.kappa*coth)
            )

    def gamma_isotensional_asymptotic_reduced(self, eta):
        r"""The reduced asymptotic approach for the
        isotensional :math:`\gamma(\eta)`.

        This function provides the reduced asymptotic approach for
        obtaining :math:`\gamma(\eta)` in the isotensional ensemble,
        a composite approximation for all :math:`\eta`
        that is valid for :math:`\varepsilon\gg 1`,

        .. math::
            \gamma(\eta) \sim \mathcal{L}(\eta) + \Delta\lambda(\eta).

        The simplification involves neglecting terms that contribute negligibly
        apart when :math:`\eta=\mathrm{ord}(\varepsilon)`, a region that tends
        to vanish for :math:`\varepsilon\gg 1`.
        This simplification is then asymptotically valid in the same limit that
        made the original asymptotic approximation possible.
        Further, this asymptotic approximation is equivalent to asymptotically
        matching the low-force entropic mechanical response
        :math:`\mathcal{L}(\eta)` to the high-force enthalpic
        mechanical response :math:`\lambda(\eta)` :cite:`buche2021chain`.

        Args:
            eta (array_like): The nondimensional force(s).

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Example:
            Verify that either asymptotic approximation obtains close-to-exact
            end-to-end lengths over many nondimensional forces
            when the nondimensional energy scale is high:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> model = uFJC(potential='morse', varepsilon=888)
                >>> eta_trial = np.logspace(-2, 1, 88)
                >>> gamma_quad = model.gamma(eta_trial, approach='quadrature')
                >>> gamma_asymp = model.gamma(eta_trial, approach='asymptotic')
                >>> (np.abs(gamma_quad-gamma_asymp)/gamma_quad < 1e-2).all()
                True
                >>> gamma_simpl = model.gamma(eta_trial, approach='asymptotic')
                >>> (np.abs(gamma_quad-gamma_simpl)/gamma_quad < 1e-2).all()
                True

        """
        return self.langevin(eta) + self.delta_lambda(eta)

    def gamma_isotensional_exact(self, eta):
        r"""Exact approach for the
        isotensional :math:`\gamma(\eta)`.

        This function provides the exact, analytical approach for obtaining
        :math:`\gamma(\eta)` in the isotensional ensemble.

        Args:
            eta (array_like): The nondimensional force(s).

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Warning:
            This is currently only available for harmonic link potentials,
            and will therefore return nan for non-harmonic potentials:

                >>> from ufjc import uFJC
                >>> uFJC().gamma([88e-2, 8], approach='exact')
                array([0.29518908, 0.97632595])
                >>> uFJC(potential='morse').gamma([88e-2, 8], approach='exact')
                array([nan, nan])

        """
        if self.potential == 'harmonic':
            return self.efjc_gamma_isotensional_exact(eta)
        else:
            return np.nan*eta

    def efjc_gamma_isotensional_exact(self, eta):
        r"""The exact approach for the EFJC
        isotensional :math:`\gamma(\eta)`.

        This function provides the exact, analytical approach for obtaining
        :math:`\gamma(\eta)` in the isotensional ensemble
        for harmonic link potentials, i.e. for the EFJC model,

        .. math::
            \gamma(\eta) =
            \mathcal{L}(\eta) +
            \frac{\eta}{\kappa}\left[1 +
            \frac{1 - \mathcal{L}(\eta)\coth(\eta)}
            {1 + (\eta/\kappa)\coth(\eta)}\right]
            + \frac{\partial}{\partial\eta}\,\ln\left[1+g(\eta)\right]
            ,

        where the function :math:`g(\eta)` is given by :cite:`buche2022freely`

        .. math::
            g(\eta) =
            \frac{
            e^{\eta}\left(\frac{\eta}{\kappa} + 1\right)
                \,\mathrm{erf}\left(\frac{\eta+\kappa}{\sqrt{2\kappa}}\right)
            - e^{-\eta}\left(\frac{\eta}{\kappa} - 1\right)
                \,\mathrm{erf}\left(\frac{\eta-\kappa}{\sqrt{2\kappa}}\right)}
            {4\sinh(\eta)\left[1 + (\eta/\kappa)\coth(\eta)\right]}
            - \frac{1}{2}
            .

        Args:
            eta (array_like): The nondimensional force(s).

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Example:
            Compute the exact nondimensional end-to-end length of the EFJC
            in the isotensional ensemble under a nondimensional force of 23:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='harmonic')
                >>> model.gamma(23, approach='exact')
                array([1.22689438])

        """
        # Avoid overflow
        eta[eta > self.maximum_exponent] = self.maximum_exponent
        eta[eta == 0] = 1e-88
        eta[eta < 1e-88] = 1e-88

        # Pre-compute some common terms
        ep = erf((eta + self.kappa)/np.sqrt(2*self.kappa))
        en = erf((eta - self.kappa)/np.sqrt(2*self.kappa))
        tp = np.exp(eta)*(eta/self.kappa + 1)*ep
        tn = np.exp(-eta)*(eta/self.kappa - 1)*en
        tc = tp - tn
        coth = self.coth(eta)
        csch = coth/np.cosh(eta)

        # Compute the numerator and denominator of the complicated term
        denominator = 1 + eta/self.kappa*coth + tc/np.sinh(eta)/2
        numerator = (coth - eta*csch**2)/self.kappa
        numerator += -coth*csch*tc/2
        numerator += csch*np.exp(eta)*(eta/self.kappa + 1 + 1/self.kappa)*ep/2
        numerator += csch*np.exp(-eta)*(eta/self.kappa - 1 - 1/self.kappa)*en/2
        numerator += csch*np.sqrt(2/self.kappa/np.pi) * \
            np.exp(-(eta**2 + self.kappa**2)/2/self.kappa)

        # Return the exact result, making use of existing relations
        gamma_reduced = self.gamma_isotensional_asymptotic_reduced(eta)
        return gamma_reduced + numerator/denominator

    def gamma_isotensional_quadrature(self, eta):
        r"""The numerical quadrature approach for the
        isotensional :math:`\gamma(\eta)`.

        Args:
            eta (array_like): The nondimensional force(s).

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Example:
            Compute the nondimensional end-to-end length of the EFJC model
            at many nondimensional forces using the numerical quadrature
            approach, ensuring the exact result is always obtained:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> model = uFJC()
                >>> eta_trial = np.logspace(-2, 2, 88)
                >>> gamma_quad = model.gamma(eta_trial, approach='quadrature')
                >>> gamma_exact = model.gamma(eta_trial, approach='exact')
                >>> np.isclose(gamma_quad, gamma_exact).all()
                True

        """
        # Restrict the upper limit of integration if the link can break
        if hasattr(self, 'lambda_max'):
            upper_lim = self.lambda_max
        else:
            upper_lim = np.inf

        # Compute the rescaled partition function for a single link
        def z_fun(lambda_):
            expo_1 = eta*lambda_ - self.varepsilon*self.phi(lambda_) \
                + np.log(lambda_) - np.log(eta)
            expo_2 = expo_1 - 2*eta*lambda_
            if (np.array([expo_1, expo_2]) < self.maximum_exponent).all():
                return np.exp(expo_1) - np.exp(expo_2)
        z_rescaled = quad_vec(z_fun, self.minimum_float, upper_lim)[0]

        # Compute and return the nondimensional end-to-end length gamma
        def gamma_fun(lambda_):
            expo_1 = eta*lambda_ - self.varepsilon*self.phi(lambda_) \
                + 2*np.log(lambda_) - np.log(eta)
            expo_2 = expo_1 - 2*eta*lambda_
            expo_3 = expo_1 - np.log(lambda_) - np.log(eta)
            expo_4 = expo_3 - 2*eta*lambda_
            expos = np.array([expo_1, expo_2, expo_3, expo_4])
            if (expos < self.maximum_exponent).all():
                term_1 = np.exp(expo_1) + np.exp(expo_2)
                term_2 = np.exp(expo_3) - np.exp(expo_4)
                return (term_1 - term_2)/z_rescaled
        return quad_vec(gamma_fun, self.minimum_float, upper_lim)[0]

    def gamma_isotensional_MHMCMC(self, eta, **kwargs):
        r"""The Monte Carlo approach for the isotensional :math:`\gamma(\eta)`.

        Args:
            eta (array_like): The nondimensional force(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to ``MHMCMC.gamma_isotensional_MHMCMC``.

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Example:
            Compute the nondimensional end-to-end length of the log-squared-FJC
            in the isotensional ensemble under a nondimensional force of 5
            using the Monte Carlo approach and ensure that it is within
            ten percent of the numerical quadrature result:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='log-squared')
                >>> gamma_MHMCMC = model.gamma(5, approach='monte carlo')
                >>> gamma_quad = model.gamma(5, approach='quadrature')
                >>> np.abs(gamma_quad - gamma_MHMCMC)/gamma_quad < 1e-1
                array([ True])

        """
        return MHMCMC(self).gamma_isotensional_MHMCMC(eta, **kwargs)

    def eta_isotensional(self, gamma, **kwargs):
        r"""Invert the isotensional :math:`\gamma(\eta)`
        for the isotensional :math:`\eta(\gamma)`.

        This function obtains the isotensional nondimensional
        single-chain mechanical response :math:`\eta(\gamma)`
        by inverting the isotensional :math:`\gamma(\eta)`.

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to ``gamma_isotensional``.

        Returns:
            numpy.ndarray: The nondimensional force(s).

        Example:
            Check that :math:`\eta[\gamma(\eta)] = \eta\,`:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> model = uFJC(potential='morse')
                >>> def check_gamma(eta):
                ...     gamma_fun = lambda eta: model.gamma_isotensional(eta)
                ...     eta_fun = lambda gamma: model.eta_isotensional(gamma)
                ...     return np.isclose(eta_fun(gamma_fun(eta))[0], eta)
                >>> check_gamma(np.random.rand())
                True

        """
        def gamma_fun(eta):
            return self.gamma_isotensional(eta, **kwargs)
        return self.inv_fun_1D(gamma, gamma_fun, power=4, maxiter=250)

    def beta_Pi_config(self, config, eta_vector):
        r"""The nondimensional total potential energy of a configuration.

        This function provides the nondimensional total potential energy
        :math:`\beta \Pi = \beta U - N_b\eta\gamma`
        given the configuration of the chain, i.e. the
        vector position of each atom/hinge relative to the first one.

        Args:
            config (numpy.ndarray): The configuration of the chain,
                a :math:`(N_b+1)`-by-3 numpy array.

        Returns:
            float: The nondimensional total potential energy :math:`\beta\Pi`.

        """
        beta_U = 0
        for j in range(1, len(config)):
            lambda_ = la.norm(config[j, :] - config[j - 1, :])
            beta_U += self.beta_u(lambda_)
        return beta_U - eta_vector.dot(config[-1, :] - config[0, :])



"""An example module comparing isotensional approaches.

This module contains a ``main`` function that allows the comparison of
different approaches of obtaining the single-chain mechanical response of
a given single-chain model in the isotensional ensemble.
The results are plotted using ``matplotlib``.
When executed at the command line, flags are passed as keyword arguments.

Example:
    Compare approaches, using 1000 samples for the Monte Carlo approach:

    ::

        python -m ufjc.examples.approaches --num_samples 1000

"""


def main(**kwargs):
    r"""Main function for the module.

    This is the main function, called when executing the module from the
    command line, also available when importing the module.

    Args:
        **kwargs: Arbitrary keyword arguments.
            Passed to ``uFJC`` instantiation.

    Example:

        .. plot::

            Plot and compare each approach for each available potential,
            including examples of polynomial and fully custom potentials,
            for :math:`\varepsilon=88`, keeping the harmonic result (EFJC)
            as a comparison for non-harmonic cases:

                >>> import numpy as np
                >>> from ufjc.examples import approaches
                >>> approaches.main(potential='harmonic')
                >>> approaches.main(potential='log-squared')
                >>> approaches.main(potential='morse')
                >>> approaches.main(potential='lennard-jones')
                >>> approaches.main(potential='mie', n=10, m=4)
                >>> approaches.main(potential='polynomial',
                ...     coefficients=[1, 2, 3])
                >>> approaches.main(potential='custom',
                ...     varepsilon=88,
                ...     phi=lambda lambda_: 1 - np.cos(lambda_ - 1),
                ...     eta_link=lambda lambda_: 88*np.sin(lambda_ - 1),
                ...     delta_lambda=lambda eta: np.arcsin(eta/88),
                ...     kappa=88,
                ...     c=2/3,
                ...     lambda_max=1 + np.pi/2,
                ...     eta_max=88)

    """

    # Import external modules
    import numpy as np
    import matplotlib.pyplot as plt

    # Import internal modules
    from ufjc import uFJC

    # Create the single-chain model
    model = uFJC(**kwargs)

    # Nondimensional forces to apply
    if kwargs.get('eta_max', None) is not None:
        eta_max = kwargs.get('eta_max', None)
    elif hasattr(model, 'eta_max'):
        eta_max = 0.95*model.eta_max
    else:
        eta_max = (0.5 + model.kappa/4)
    eta = np.linspace(1e-2, eta_max, int(kwargs.get('num_points_eta', 100)))
    eta_mc = np.linspace(eta[0], eta[-1], int(kwargs.get('num_forces', 5)))

    # Plot each available approach for the isotensional gamma(eta)
    plt.figure()
    if model.potential == 'harmonic':
        plt.plot(model.gamma(eta, approach='exact'), eta,
                 label='exact', linewidth=2.5)
    else:
        model_h = uFJC(varepsilon=model.kappa, potential='harmonic')
        gamma_h = model_h.gamma(eta, approach='exact')
        plt.plot(gamma_h, eta, 'k-', label='(EFJC exact)', linewidth=1.5)
    plt.plot(model.gamma(eta, approach='asymptotic'), eta,
             '--', label='asymptotic', linewidth=2.5)
    plt.plot(model.gamma(eta, approach='reduced'), eta,
             '--', label='reduced', linewidth=2.5)
    plt.plot(model.gamma(eta, approach='quadrature'), eta,
             ':', label='quadrature', linewidth=4)
    plt.plot(model.gamma(eta_mc, approach='monte carlo', **kwargs), eta_mc,
             'ko', label='Monte Carlo')
    plt.legend()
    plt.xlabel(r'$\gamma$')
    plt.ylabel(r'$\eta$')
    plt.xlim([-0.03, 1.08*model.gamma(eta[-1], approach='asymptotic')])
    plt.title(model.potential + '-FJC isotensional ' + r'$\gamma(\eta)$')
    plt.show()


if __name__ == '__main__':  # pragma: no cover

    # Import external modules
    import argparse

    # Parse args
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--potential', type=str, default='harmonic')

    # Parse arbitrary float args
    for arg in parser.parse_known_args()[1]:
        if arg.startswith('--'):
            parser.add_argument(arg.split('=')[0], type=float)

    # Execute main function with passed args as kwargs
    main(**vars(parser.parse_args()))



"""The core module for the uFJC single-chain model.

This module consist of the class ``uFJC`` which, upon instantiation,
becomes a uFJC single-chain model instance with methods for computing
single-chain quantities in either thermodynamic ensemble.

Example:
    Import and create an instance of the model:

        >>> from ufjc import uFJC
        >>> model = uFJC()

"""

# Import internal modules
from .isometric import uFJCIsometric
from .potential import Potential

# Import external modules
import numpy as np
from scipy.integrate import quad


class uFJC(Potential, uFJCIsometric):
    r"""The uFJC single-chain model class.

    This class represents the uFJC single-chain model, meaning that an
    instance of this class is a uFJC single-chain model instance.
    It inherits all attributes and methods from the ``uFJCIsometric``
    class, which inherits all attributes and methods from the
    ``uFJCIsotensional`` class, which inherits all attributes and
    methods from the ``BasicUtility`` class.
    It also inherits a potential from the ``Potential`` class
    as the attribute ``pot``, a model instance itself.
    Keyword arguments are utilized during instantiation in order to
    specify model parameters; see the inherited classes and various examples
    through the documentation for more information.

    Attributes:
        N_b (int): The number of links in the chain.
        init_config (np.ndarray): The initial configuration for Monte Carlo.
        c_kappa (float): The constant used for ideal/Gaussian approximations.
        P_eq_normalizations (dict): The normalizations for the
            nondimensional equilibrium distribution ``P_eq``
            for each approach.
        pot (object): The link potential model instance.

    Note:
        The attributes of a instantiated model should not be changed.
        There are several quantities that are computed and stored during
        instantiation that depend on these attributes and are not re-computed.

    """
    def __init__(self, **kwargs):
        """Initializes the ``uFJC`` class,
        producing a uFJC single-chain model instance.

        First, initialize and inherit all attributes and methods
        from ``uFJCIsometric`` and ``Potential`` class instances;
        the latter assigns a potential (as attributes) to the model instance.
        Next, assign the attributes listed above.
        Then assign a default incremental link stretch function
        ``delta_lambda`` (inversion of ``eta_link``) if not already
        available from the potential choice.
        Also, adjust both ``delta_lambda`` and ``beta_u`` above
        ``eta_max`` and ``lambda_max``, if they exist for the potential,
        to prevent calculation errors.
        Finally, compute and store the normalization for ``P_eq``.

        Args:
            **kwargs: Arbitrary keyword arguments.

        """
        uFJCIsometric.__init__(self)
        Potential.__init__(self, **kwargs)

        # Get general single-chain model parameters
        self.N_b = int(kwargs.get('N_b', 1))
        self.P_eq_normalizations = {}

        # Default initial configuration for Monte Carlo calculations
        self.init_config = np.append(
            np.array([np.arange(self.N_b + 1)]).T,
            np.zeros((self.N_b + 1, 2)),
            axis=1
        )

        # Constant for ideal/Gaussian approximations
        self.c_kappa = self.kappa*(self.kappa + 1) / \
            (self.kappa**2 + 6*self.kappa + 3)

        # Default to inverting eta(lambda) for the link if unavailable
        if hasattr(self, 'delta_lambda') is False:
            self.delta_lambda = lambda eta: self.np_array(
               self.inv_fun_1D(eta, self.eta_link, guess=1, maxiter=250) - 1
            )

        # Prevent calculating bond stretch above a possible maximum force
        if hasattr(self, 'eta_max'):
            self.__old_delta_lambda = self.delta_lambda

            # Returns a high number above eta_max rather than calculating
            def delta_lambda_safe(eta):
                eta = self.np_array(eta)
                indices = eta > self.eta_max
                eta[indices] = 0
                delta_lambda = self.__old_delta_lambda(eta)
                delta_lambda[indices] = self.maximum_exponent
                return delta_lambda

            # Replace the original function with this safe function
            self.delta_lambda = delta_lambda_safe

        # Inhibit Monte Carlo sampling link stretches above lambda_max
        if hasattr(self, 'lambda_max'):
            self.__old_beta_u = self.beta_u

            # Returns a prohibitively high number above lambda_max
            def beta_u_safe(lambda_):
                return self.__old_beta_u(lambda_) + \
                    self.maximum_exponent * \
                    np.heaviside(lambda_ - self.lambda_max, 0)

                # Replace the original function with this safe function
            self.beta_u = beta_u_safe

    def gamma(self, eta, **kwargs):
        r"""The nondimensional end-to-end length.

        This function computes the scalar nondimensional end-to-end length as a
        function of the scalar nondimensional force, :math:`\gamma(\eta)`.
        In the isotensional ensemble, it is given by
        :cite:`fiasconaro2019analytical`

        .. math::
            \gamma(\eta) = \frac{\partial}{\partial\eta}\,\ln\mathfrak{z}(\eta)
            ,

        where the single-link (since :math:`\gamma(\eta)` is independent of
        :math:`N_b` in the isotensional ensemble for the uFJC model)
        nondimensional isotensional partition function is given by
        :cite:`buche2022freely`

        .. math::
            \mathfrak{z}(\eta) =
            4\pi\int \frac{\sinh(\lambda\eta)}{\lambda\eta} \,
            e^{-\varepsilon\phi(\lambda)} \, \lambda^2 d\lambda
            .

        This function is rarely exactly
        analytically available, is sometimes accurately approximated using
        analytically tractable asymptotic relations, and is otherwise
        numerically calculated using Monte Carlo or quadrature approaches.
        For the uFJC model, accurate asymptotic relations are available,
        and for the EFJC model (harmonic link potentials), it is actually
        possible to exactly find :math:`\gamma(\eta)` closed-form
        :cite:`balabaev2009extension`.

        Args:
            eta (array_like): The nondimensional force(s).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Examples:
            Create an EFJC model with a nondimensional link energy
            :math:`\varepsilon=23`, and calculate the nondimensional
            end-to-end length under an nondimensional force of 8
            using many of the available approaches:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='harmonic', varepsilon=23)
                >>> model.gamma(8, approach='exact')
                array([1.25508427])
                >>> model.gamma(8, approach='asymptotic')
                array([1.25508427])
                >>> model.gamma(8, approach='reduced')
                array([1.22282631])
                >>> model.gamma(8, approach='quadrature')
                array([1.25508427])

        Example:
            Verify that the ideal approximation is valid
            for small nondimensional forces:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> model = uFJC()
                >>> gamma_exact = model.gamma(1e-2, approach='exact')
                >>> gamma_ideal = model.gamma(1e-2, ideal=True)
                >>> np.isclose(gamma_exact, gamma_ideal)
                array([ True])

        Example:
            Verify that the single-chain mechanical response is the same
            in the isotensional ensemble as it is in the isometric ensemble
            when utilizing the Legendre transformation method:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='harmonic', varepsilon=23)
                >>> eta = np.random.rand(88)
                >>> (np.abs(model.gamma(eta, ensemble='isotensional') -
                ...         model.gamma(eta, ensemble='isometric',
                ...                     method='legendre')) < 1e-6).all()
                True

        """
        if kwargs.get('ideal', False) is True:
            return self.np_array(eta)/(3*self.c_kappa)
        else:
            if kwargs.get('ensemble', 'isotensional') == 'isometric':
                return self.gamma_isometric(eta, **kwargs)
            else:
                return self.gamma_isotensional(eta, **kwargs)

    def eta(self, gamma, **kwargs):
        r"""The nondimensional force.

        This function computes the scalar nondimensional force as a function of
        the scalar nondimensional end-to-end length, :math:`\eta(\gamma)`.
        In the isometric ensemble, it is given by :cite:`manca2012elasticity`

        .. math::
            \gamma(\eta) = -\frac{1}{N_b}
            \frac{\partial}{\partial\gamma}
            \,\ln\mathfrak{Q}(\boldsymbol{\gamma})
            ,

        where the nondimensional isometric partition function is given by
        :cite:`buche2021fundamental`

        .. math::
            \mathfrak{Q}(\boldsymbol{\gamma}) =
            \iiint d^3\boldsymbol{\lambda}_1
            \,e^{-\varepsilon\phi(\lambda_1)} \cdots
            \iiint d^3\boldsymbol{\lambda}_{N_b}
            \,e^{-\varepsilon\phi(\lambda_{N_b})}
            ~\delta^3\left(\sum_{j=1}^{N_b}\boldsymbol{\lambda}_j
            - \boldsymbol{\gamma}\right)
            .

        This is analytically intractable for nearly all single-chain models,
        including the uFJC model. Fortunately, there are both exact numerical
        methods, such as Monte Carlo calculations, and accurate approximation
        methods, such as the Legendre transformation method which becomes
        valid relatively quickly as :math:`N_b` increases.

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments.

        Returns:
            numpy.ndarray: The nondimensional force(s).

        Example:
            Compute the nondimensional force at an end-to-end length in the
            isometric ensemble, which (by default) is done using the Legendre
            transformation method and the asymptotic approach to get
            :math:`\gamma(\eta)` in the isotensional ensemble. Using the
            Legendre transformation method with a given isotensional approach
            for the isometric ensemble is equivalent to using the same approach
            in the isotensional ensemble:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='harmonic', varepsilon=23, N_b=8)
                >>> model.eta(1.25508427, ensemble='isometric')
                array([8.00000001])
                >>> model.eta(1.25508427, ensemble='isotensional')
                array([8.00000001])

        Example:
            Verify that the ideal approximation is valid for small
            nondimensional end-to-end lengths:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> model = uFJC()
                >>> eta_exact = model.eta(1e-2, approach='exact')
                >>> eta_ideal = model.eta(1e-2, ideal=True)
                >>> np.abs((eta_exact - eta_ideal)/eta_exact) < 1e-3
                array([ True])

        """
        if kwargs.get('ideal', False) is True:
            return 3*self.c_kappa*self.np_array(gamma)
        else:
            if kwargs.get('ensemble', 'isotensional') == 'isometric':
                return self.eta_isometric(gamma, **kwargs)
            else:
                return self.eta_isotensional(gamma, **kwargs)

    def vartheta(self, gamma, **kwargs):
        r"""The nondimensional Helmholtz free energy per link.

        This function compute the nondimensional Helmholtz free energy per link
        in the isometric ensemble, which is generally given by

        .. math::
            \vartheta(\boldsymbol{\gamma}) =
            -\ln\left[\mathfrak{Q}(\boldsymbol{\gamma})\right]^{1/N_b}
            .

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to ``vartheta_isometric``.

        Returns:
            numpy.ndarray: The nondimensional Helmholtz
            free energy(s) per link.

        Example:
            Create a uFJC single-chain model with log-squared link potentials
            and calculate the Helmholtz free energy per link at
            a nondimensional end-to-end length of a half:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='log-squared')
                >>> model.vartheta(0.5, method='legendre', approach='reduced')
                array([0.39097337])

        Example:
            Verify that the ideal approximation is valid for small
            nondimensional end-to-end lengths:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> model = uFJC()
                >>> vartheta = model.vartheta(1e-2, approach='reduced')
                >>> vartheta_ideal = model.vartheta(1e-2, ideal=True)
                >>> np.abs((vartheta - vartheta_ideal)/vartheta) < 1e-1
                array([ True])

        """
        if kwargs.get('ideal', False) is True:
            return 1.5*self.c_kappa*self.np_array(gamma)**2
        else:
            return self.vartheta_isometric(gamma, **kwargs)

    def P_eq(self, gamma, **kwargs):
        r"""The nondimensional equilibrium distribution function.

        This function computes the nondimensional probability density
        distribution of nondimensional end-to-end vectors at equilibrium,

        .. math::
            \mathscr{P}_\mathrm{eq}(\boldsymbol{\gamma}) =
            \frac{e^{-N_b\,\vartheta(\boldsymbol{\gamma})}}
            {\iiint e^{-N_b\,\vartheta(\tilde{\boldsymbol{\gamma}})}
            \,d^3\tilde{\boldsymbol{\gamma}}}
            .

        Since the uFJC model is spherically-symmetric in
        :math:`\boldsymbol{\gamma}`,  ``P_eq`` is only a function
        of the scalar nondimensional end-to-end length :math:`\gamma`,
        and the normalization can be calculated
        using a one-dimensional integral over :math:`\gamma` instead
        :cite:`buche2020statistical`,

        .. math::
            \iiint e^{-N_b\,\vartheta(\tilde{\boldsymbol{\gamma}})}
            \,d^3\tilde{\boldsymbol{\gamma}} =
            4\pi\int e^{-N_b\,\vartheta(\tilde{\gamma})}
            \,\tilde{\gamma}^2 d\tilde{\gamma}
            .

        The normalization for the distribution is computed for a given approach
        once the function is called using that approach.

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments. Passed to ``vartheta``.

        Returns:
            numpy.ndarray: The nondimensional equilibrium
            probability density(s).

        Example:
            Evaluate the probability density at an end-to-end length of 0.23
            using two different approximation methods:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='morse', N_b=8)
                >>> model.P_eq(23e-2)
                array([4.19686303])
                >>> model.P_eq(23e-2, gaussian=True)
                array([3.86142625])

        Example:
            Verify that the probability density is normalized:

                >>> from ufjc import uFJC
                >>> model = uFJC()
                >>> from scipy.integrate import quad
                >>> integrand = lambda gamma: \
                ...     4*np.pi*gamma**2*model.P_eq(gamma)
                >>> P_tot_eq = quad(integrand, 0, np.inf)[0]
                >>> np.isclose(P_tot_eq, 1)
                True

        """
        if kwargs.get('gaussian', False) is True:
            a = 1.5*self.N_b*self.c_kappa
            return (a/np.pi)**(1.5)*np.exp(-a*self.np_array(gamma)**2)
        else:

            # A variable string based on the approach
            var_str = 'P_eq_normalization_' + \
                str(kwargs.get('approach', 'asymptotic'))

            # Compute the normalization if not done already for the approach
            if self.P_eq_normalizations.get(var_str) is None:

                # Set upper limit of integration
                if hasattr(self, 'lambda_max') is True:
                    upper_lim = self.lambda_max
                else:
                    upper_lim = np.inf

                # Compute the normalzation and attribute it
                self.P_eq_normalizations[var_str] = \
                    quad(lambda gamma: 4*np.pi*gamma**2 *
                         np.exp(-self.N_b*self.vartheta(gamma)),
                         self.minimum_float,
                         upper_lim
                         )[0]

            # Return the nondimensional equilibrium probability density
            return np.exp(-self.N_b*self.vartheta(gamma, **kwargs)) / \
                self.P_eq_normalizations[var_str]

    def g_eq(self, gamma, **kwargs):
        r"""The nondimensional equilibrium radial distribution function.

        This function computes the nondimensional radial probability density
        distribution of nondimensional end-to-end lengths at equilibrium,

        .. math::
            \mathscr{g}_\mathrm{eq}(\gamma) =
            4\pi\gamma^2 \mathscr{P}_\mathrm{eq}(\gamma)
            .

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments. Passed to ``P_eq``.

        Returns:
            numpy.ndarray: The nondimensional equilibrium radial
            probability density(s).

        Example:
            Verify that the function is normalized by integrating over all
            permissible end-to-end lengths, which in effect ensures that the
            total probability is unity:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> model = uFJC(potential='morse')
                >>> from scipy.integrate import quad
                >>> P_tot_eq = quad(model.g_eq, 0, model.lambda_max)[0]
                >>> np.isclose(P_tot_eq, 1)
                True

        """
        return 4*np.pi*self.np_array(gamma)**2 * \
            self.P_eq(gamma, **kwargs)

    def k(self, gamma):
        r"""The nondimensional net forward reaction rate coefficient function.

        This function computes the net forward reaction rate coefficient
        as a function of the nondimensional end-to-end length,
        i.e. :math:`k(\gamma)=N_bk'(\gamma)`, where :math:`k'(\gamma)` is the
        forward reaction rate coefficient function for a single link.
        This function is obtained using the axioms of transition state theory
        :cite:`zwanzig2001nonequilibrium`,
        and is currently implemented to use the Legendre transformation method
        and the reduced asymptotic approach
        :cite:`buche2021chain`.

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).

        Returns:
            numpy.ndarray:
                The nondimensional net forward reaction rate coefficient(s).

        Example:
            Compute the nondimensional net forward reaction rate coefficients
            at several nondimensional end-to-end lengths:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='morse')
                >>> model.k([0.8, 1.01, 1.6])
                array([6.23755528e-01, 9.80472713e-01, 6.59105919e+07])

        """
        # Compute the corresponding force and link stretch
        eta = self.np_array(self.eta(gamma))
        lambda_ = 1 + self.delta_lambda(eta)

        # Compute the free energy barrier for a link, translated so k(0)=1
        beta_delta_Psi_TS_single_link = \
            self.varepsilon*(self.phi(1) - self.phi(lambda_))
        if hasattr(self, 'lambda_max'):
            beta_delta_Psi_TS_single_link += -self.log_over_sinh(eta) + \
                self.log_over_sinh(self.lambda_max*eta) + \
                self.lambda_max*eta*self.coth(self.lambda_max*eta) + \
                - eta*self.coth(eta)

        # Avoid overflow when returning the results
        exponent = np.log(self.N_b) - beta_delta_Psi_TS_single_link
        exponent[exponent > self.maximum_exponent] = self.maximum_exponent
        return np.exp(exponent)



"""A module for the Metropolis-Hastings Markov chain Monte Carlo method.

This module consist of the class ``MHMCMC`` which contains
attributes and methods that allow single-chain quantities
to be computed using the Metropolis-Hastings Markov chain Monte Carlo method.

Example:
    Import and instantiate the class with a single-chain model:

        >>> from ufjc import uFJC
        >>> from ufjc.monte_carlo import MHMCMC
        >>> class_instance = MHMCMC(uFJC())

"""

# Import internal modules
from .utility import BasicUtility

# Import external modules
import numpy as np
import numpy.linalg as la
from sys import platform
if platform == "linux" or platform == "linux2":
    import multiprocessing as mp


class MHMCMC(BasicUtility):
    """The Metropolis-Hastings Markov chain Monte Carlo class.

    This class contains attributes and methods used to compute
    statistical thermodynamic quantities using the
    Metropolis-Hastings Markov chain Monte Carlo method
    :cite:`haile1992molecular` for a given single-chain model.

    Attributes:
        N_b:
            The number of chain links.
            Inherited from ``single_chain_model``.
        beta_U_config:
            The configurational potential energy.
            Inherited from ``single_chain_model``.
        beta_Pi_config:
            The configurational total potential energy.
            Inherited from ``single_chain_model``.
        init_config:
            The default initial configuration.
            Inherited from ``single_chain_model``.

    """
    def __init__(self, single_chain_model):
        """Initializes the ``MHMCMC`` class.

        Initialize and inherit all attributes and methods
        from a ``BasicUtility`` class instance, and certain
        attributes and methods from the given single-chain model.

        Args:
            single_chain_model(object): The single-chain model instance.

        """
        BasicUtility.__init__(self)

        # Inherit necessary attributes from the single-chain model
        self.N_b = single_chain_model.N_b
        self.init_config = single_chain_model.init_config
        self.beta_U_config = single_chain_model.beta_U_config
        self.beta_Pi_config = single_chain_model.beta_Pi_config

    def trial_config(self, prev_config, **kwargs):
        """Generates trial configurations given a previous configuration.

        This function returns a trial configuration given a previous starting
        configuration.
        The possibility of rigid-body translation is prevented
        by fixing the first coordinate at the origin.
        In the isometric ensemble, the last coordinate is left unchanged
        in order to keep the end-to-end vector fixed.

        Args:
            prev_config (numpy.ndarray): The previous configuration.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            numpy.ndarray: The generated trial configuration.

        Example:
            Generate a trial configuration in the isometric ensemble and
            ensure that the end-to-end remains unchanged:

                >>> from ufjc import uFJC
                >>> from ufjc.monte_carlo import MHMCMC
                >>> model = uFJC(N_b=8)
                >>> prev_config = model.init_config
                >>> trial_config = MHMCMC(model).trial_config(prev_config,
                ...     ensemble='isometric')
                >>> (prev_config == trial_config)[(0,-1),:]
                array([[ True,  True,  True],
                       [ True,  True,  True]])

        """
        cov_config = kwargs.get('transition_variance', 1e-2)*np.eye(3)
        delta_config = np.random.multivariate_normal(
            np.zeros(3),
            cov_config,
            len(prev_config)
        )
        delta_config[0, :] = 0
        if kwargs.get('ensemble', 'isotensional') == 'isometric':
            delta_config[-1, :] = 0
        return prev_config + delta_config

    def mh_next_config(self, prev_config, **kwargs):
        """Generates accepted configurations given a previous configuration.

        This function returns an accepted configuration given a
        previous configuration. Trial configurations are generated and
        repeatedly rejected until accepted as the next configuration
        based on the Metropolis-Hastings algorithm.

        Args:
            prev_config (numpy.ndarray): The previous configuration.
            **kwargs: Arbitrary keyword arguments.
                Passed to ``trial_config``.

        Returns:
            numpy.ndarray: The accepted (next) configuration.

        """
        next_config = None
        urng = kwargs.get('urng', np.random.uniform)
        while next_config is None:
            trial_config = self.trial_config(prev_config, **kwargs)
            delta_bEc = self.__bEc(trial_config) - self.__bEc(prev_config)
            if (delta_bEc < 0) or (np.exp(-delta_bEc) > urng()):
                next_config = trial_config
        return next_config

    def mh_next_config_append(self, prev_configs, **kwargs):
        """Append next configuration onto a history of previous configurations.

        This function calls ``mh_next_config`` and appends the resulting next
        configuration onto a given history of previous configurations.

        Args:
            prev_configs (numpy.ndarray): The previous
                history of configurations.
            **kwargs: Arbitrary keyword arguments.
                Passed to ``mh_next_config``.

        Returns:
            numpy.ndarray: The updated history of configurations.

        """
        return np.append(
            prev_configs,
            np.array([self.mh_next_config(prev_configs[-1], **kwargs)]),
            axis=0)

    def burn_in(self, init_config, **kwargs):
        """Burn-in (throw away samples) until in the effective sampling region.

        This function runs a Monte Carlo calculation without utilizing samples,
        instead throwing them away until a desired number have been thrown away
        or a desired tolerance has been reached.
        This is typically done to obtain a burned-in configuration to use as
        the initial configuration for an actual Monte Carlo calculation.

        Args:
            init_config (numpy.ndarray):
                The initial configuration.
            **kwargs: Arbitrary keyword arguments.
                Passed to ``mh_next_config_append``.

        Returns:
            numpy.ndarray: The final, burned-in configuration.

        """
        config_history = np.array([init_config])
        prev_rmc = init_config

        # Generate new configurations until certain criterion are met
        tol = kwargs.get('tol', np.inf)
        num_samples_burn_in = int(kwargs.get('num_samples_burn_in', int(1e3)))
        n_res = tol + 1
        while (n_res > tol) or (len(config_history) < num_samples_burn_in):

            # Generate an accepted configuration
            config_history = self.mh_next_config_append(
                config_history, **kwargs)

            # Compute the normalized residual in mean configuration
            next_rmc = np.mean(config_history, axis=0)
            n_res = la.norm(next_rmc - prev_rmc)/la.norm(prev_rmc)
            prev_rmc = next_rmc

        # Return the configuration at the end of the burn-in period
        return config_history[-1]

    def parallel_calculation(self, serial_fun, **kwargs):
        """Monte Carlo calculation averaged over several parallel processes.

        This function performs several Monte Carlo calculations in parallel
        and returns the average result from all processes.
        Each Monte Carlo calculation is performed using ``serial_fun``.
        The default is to utilize all available processors.

        Args:
            serial_fun(function):
                The function for a single Monte Carlo calculation.
            **kwargs: Arbitrary keyword arguments.
                Passed to ``serial_fun``.

        Returns:
            float: The process-averaged result.

        Note:
            To avoid issues with ``multiprocessing`` and MacOS/Windows,
            ``num_processes`` is set to 1 for anything but Linux.

        """
        burned_in_config = self.burn_in(self.init_config, **kwargs)
        num_processes = 1
        if platform == "linux" or platform == "linux2":
            num_processes = kwargs.get('num_processes', mp.cpu_count())
        if num_processes > 1:

            # Create an output queue
            output = mp.Queue()

            # Set up the function run by each process
            def fun(seed, output):
                output.put(
                    serial_fun(
                        burned_in_config,
                        urng=np.random.RandomState(seed).random, **kwargs
                    )
                )

            # Create the processes, each with a random seed
            processes = [
                mp.Process(target=fun, args=(seed, output))
                for seed in np.random.randint(88, size=num_processes)
            ]

            # Run each process
            for p in processes:
                p.start()

            # Exit each process after running them all
            for p in processes:
                p.join()

            # Get the results from each process from the output queue
            process_results = [output.get() for p in processes]

            # Calculate and return the mean over all processes
            return np.mean(process_results)

        # Otherwise do a single serial calculation
        else:
            return serial_fun(burned_in_config, **kwargs)

    def gamma_isotensional_MHMCMC_serial(self, config, **kwargs):
        r"""Serial Monte Carlo calculation function for
        the isotensional :math:`\gamma(\eta)`.

        This function computes the nondimensional end-to-end length
        using a Monte Carlo calculation given an initial configuration.

        Args:
            config (numpy.ndarray):
                The initial configuration, typically already burned-in.
            **kwargs: Arbitrary keyword arguments.
                Passed to ``mh_next_config``.

        Returns:
            float: The nondimensional end-to-end length.

        Example:
            Calculate the nondimensional end-to-end length under a
            nondimensional force of 23 and ensure that it is within
            one percent of the exact result:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> from ufjc.monte_carlo import MHMCMC
                >>> gamma_MHMCMC = MHMCMC(uFJC()). \
                ...     gamma_isotensional_MHMCMC(23, num_processes=1)
                >>> gamma_exact = uFJC().gamma(23, approach='exact')
                >>> np.abs(gamma_exact - gamma_MHMCMC)/gamma_exact < 1e-2
                array([ True])

        """
        gamma_vector = 0
        num_samples = int(kwargs.get('num_samples', int(1e4)))
        for counter in range(1, 1 + num_samples):
            gamma_vector_next = (
                config[-1, :] - config[0, :]
            )/self.N_b
            gamma_vector += (
                gamma_vector_next - gamma_vector
            )/counter
            config = self.mh_next_config(config, **kwargs)

        # Return gamma as a scalar
        return la.norm(gamma_vector)

    def gamma_isotensional_MHMCMC(self, eta, **kwargs):
        r"""Parallel Monte Carlo calculation function for
        the isotensional :math:`\gamma(\eta)`.

        Given a nondimensional force, this function computes the
        nondimensional end-to-end length using the average of several
        Monte Carlo calculations run in parallel.
        Given multiple nondimensional forces, this function does the above
        in serial for each nondimensional force.

        Args:
            eta (array_like):
                The nondimensional force(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to ``parallel_calculation``.

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Example:
            Calculate the nondimensional end-to-end length under a
            nondimensional force of 23 and ensure that it is within
            one percent of the exact result:

                >>> from ufjc import uFJC
                >>> from ufjc.monte_carlo import MHMCMC
                >>> gamma_MHMCMC = MHMCMC(uFJC()).gamma_isotensional_MHMCMC(23)
                >>> gamma_exact = uFJC().gamma(23, approach='exact')
                >>> np.abs(gamma_exact - gamma_MHMCMC)/gamma_exact < 1e-2
                array([ True])

        """
        eta = self.np_array(eta)
        gamma = np.zeros(eta.shape)
        for index in range(len(eta)):

            # Assume eta is aligned with the z-axis, without loss of generality
            eta_vector = np.array([0, 0, eta[index]])

            # Use the total potential energy Pi (isotensional ensemble)
            self.__bEc = lambda config: self.beta_Pi_config(config, eta_vector)

            # Calculate gamma using an ensemble average
            gamma[index] = self.parallel_calculation(
                self.gamma_isotensional_MHMCMC_serial,
                **kwargs
            )

        # Return the resulting averaged gamma
        return gamma



"""The core module for the SWFJC single-chain model.

This module contains the core class ``SWFJC`` which, upon instantiation,
becomes a SWFJC single-chain model instance with methods for computing
single-chain quantities in either thermodynamic ensemble.
The ``SWFJCIsometric`` and ``SWFJCIsotensional`` classes are also
contained within this module.
The SWFJC is the uFJC model with a square-well link potential,
which is a special case that is efficiently treated separately
as it can be solved exactly :cite:`buche2022freely`.

Example:
    Import and create an instance of the model:

        >>> from ufjc import SWFJC
        >>> model = SWFJC()

"""

# Import internal modules
from .utility import BasicUtility

# Import external modules
import numpy as np


class SWFJCIsometric(BasicUtility):
    """The SWFJC model class for the isometric ensemble.

    Attributes:
        N (int): The number of links in the chain.
        varsigma (float): The nondimensional well width.

    """
    def __init__(self, N_b=8, varsigma=3):
        """Initializes the ``SWFJCIsometric`` class.

        Initialize and inherit all attributes and methods
        from a ``BasicUtility`` class instance.

        Args:
            N_b (int, optional, default=8):
                The number of links in the chain.
            varsigma (float, optional, default=3):
                The nondimensional well width.

        """
        super().__init__()
        self.N_b = N_b
        self.varsigma = varsigma


class SWFJCIsotensional(SWFJCIsometric):
    """The SWFJC model class for the isotensional ensemble.

    Attributes:
        N (int): The number of links in the chain.
        varsigma (float): The nondimensional well width.

    """
    def __init__(self, N_b=8, varsigma=2):
        """Initializes the ``SWFJCIsotensional`` class.

        Initialize and inherit all attributes and methods
        from a ``SWFJCIsometric`` class instance.

        Args:
            N_b (int, optional, default=8):
                The number of links in the chain.
            varsigma (float, optional, default=3):
                One plus the nondimensional well width.

        """
        super().__init__()
        self.N_b = N_b
        self.varsigma = varsigma

    def z(self, eta):
        r"""The nondimensional single-link isotensional partition function
        as a function of the nondimensional force,

        .. math::
            \mathfrak{z}(\eta) =
            \frac{1}{\eta^3}\left[
                \varsigma\eta\cosh(\varsigma\eta)
                - \sinh(\varsigma\eta)
                - \eta\cosh(\eta) + \sinh(\eta)
            \right].

        Args:
            v (array_like): The nondimensional force.

        Returns:
            numpy.ndarray: The nondimensional isotensional partition function.

        """
        eta = self.np_array(eta)
        eta_zero = eta == 0
        eta[eta_zero] = -1
        z = (
            np.sinh(eta) - eta*np.cosh(eta)
            - np.sinh(self.varsigma*eta)
            + self.varsigma*eta*np.cosh(self.varsigma*eta)
        )/eta**3
        z[eta_zero] = (self.varsigma**3 - 1)/3
        return z

    def beta_varphi(self, eta):
        r"""The nondimensional isotensional free energy as a function of
        the nondimensional force,

        .. math::
            \beta\varphi(\eta) =
            -\ln\mathfrak{z}(\eta).

        Note that this becomes the isotensional free energy
        of the FJC model as :math:`\varsigma` goes to zero,

        .. math::
            \lim_{\varsigma\to 0}\beta\varphi(\eta) =
            \ln\left[\frac{\eta}{\sinh(\eta)}\right].

        Args:
            v (array_like): The nondimensional force.

        Returns:
            numpy.ndarray: The nondimensional isotensional free energy.

        Example:
            Plot the nondimensional isotensional free energy as a function of
            the nondimensional force for varying :math:`\varsigma`:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from ufjc.swfjc import SWFJCIsotensional
                >>> eta = np.linspace(0, 10, 1000)[1:]
                >>> _ = plt.figure()
                >>> for varsigma in [2, 3, 5, 10]:
                ...     model = SWFJCIsotensional(varsigma=varsigma)
                ...     _ = plt.plot(eta, model.beta_varphi(eta),
                ...                  label=r'$\varsigma=$'+str(varsigma))
                >>> _ = plt.plot(eta, np.log(eta/np.sinh(eta)),
                ...              'k--', label='FJC')
                >>> _ = plt.xlabel(r'$\eta$')
                >>> _ = plt.ylabel(r'$\beta\varphi$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return (np.log(self.z(0)) - np.log(self.z(eta)))/self.varsigma

    def gamma_isotensional(self, eta):
        r"""The nondimensional end-to-end length as a function of
        the nondimensional force in the isotensional ensemble,

        .. math::
            \gamma(\eta) =
            \frac{\partial}{\partial\eta}\,\ln\mathfrak{z}(\eta) =
            \frac{
                \varsigma^2\eta\sinh(\varsigma\eta) - \eta\sinh(\eta)
            }{
                \varsigma\eta\cosh(\varsigma\eta)
                - \sinh(\varsigma\eta)
                - \eta\cosh(\eta) + \sinh(\eta)
            }

        Note that this becomes the Langevin function of the FJC model
        as :math:`\varsigma` goes to zero,

        .. math::
            \lim_{\varsigma\to 0}\gamma(\eta) =
            \coth(\eta) - \frac{1}{\eta} =
            \mathcal{L}(\eta).

        Args:
            v (array_like): The nondimensional force.

        Returns:
            numpy.ndarray: The nondimensional end-to-end length.

        Example:
            Plot the nondimensional single-chain mechanical response
            in the isotensional ensemble for varying :math:`\varsigma`:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from ufjc.swfjc import SWFJCIsotensional
                >>> eta = np.linspace(0, 10, 1000)[1:]
                >>> _ = plt.figure()
                >>> for varsigma in [2, 3, 5, 10]:
                ...     model = SWFJCIsotensional(varsigma=varsigma)
                ...     _ = plt.plot(
                ...         model.gamma_isotensional(eta)/varsigma,
                ...         eta, label=r'$\varsigma=$'+str(varsigma))
                >>> _ = plt.plot(1/np.tanh(eta) - 1/eta, eta,
                ...              'k--', label='FJC')
                >>> _ = plt.xlabel(r'$\gamma/\varsigma$')
                >>> _ = plt.ylabel(r'$\eta$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return (
            self.varsigma**2*eta*np.sinh(self.varsigma*eta)
            - eta*np.sinh(eta)
        )/(
            np.sinh(eta) - eta*np.cosh(eta)
            - np.sinh(self.varsigma*eta)
            + self.varsigma*eta*np.cosh(self.varsigma*eta)
        ) - 3/eta


class SWFJC(SWFJCIsotensional):
    """The SWFJC single-chain model class.

    Attributes:
        N (int): The number of links in the chain.
        varsigma (float): The nondimensional well width.

    """
    def __init__(self):
        """Initializes the ``SWFJC`` class.

        Initialize and inherit all attributes and methods
        from a ``SWFJCIsotensional`` class instance.

        Args:
            N_b (int, optional, default=8):
                The number of links in the chain.
            varsigma (float, optional, default=3):
                The nondimensional well width.

        """
        super().__init__()



"""An example module involving isotensional asymptotics.

This module contains a ``main`` function that allows the asymptotic
approaches to be compared to exact approaches, when available, and
otherwise to the numerical quadrature approach.
The results are plotted using ``matplotlib``.
When executed at the command line, flags are passed as keyword arguments.

Example:
    Compare the asymptotic approaches to the exact approach for the EFJC
    model for a few nondimensional energy scales:

    ::

        python -m ufjc.examples.asymptotics --varepsilon_list 10 25 100 1000

"""


def main(**kwargs):
    r"""Main function for the module.

    This is the main function, called when executing the module from the
    command line, also available when importing the module.

    Args:
        **kwargs: Arbitrary keyword arguments.
            Passed to ``uFJC`` instantiation.

    Example:

        .. plot::

            Compare the asymptotic approaches for many models and parameters:

                >>> from ufjc.examples import asymptotics
                >>> asymptotics.main(potential='harmonic',
                ...     varepsilon_list=[10, 25, 100, 1000])
                >>> asymptotics.main(potential='log-squared',
                ...     varepsilon_list=[10, 25, 250])
                >>> asymptotics.main(potential='morse',
                ...     varepsilon_list=[10, 25, 250])
                >>> asymptotics.main(potential='lennard-jones',
                ...     varepsilon_list=[1, 2, 5, 15])
                >>> asymptotics.main(potential='mie', n=10, m=4,
                ...     varepsilon_list=[1, 2, 5, 15])
                >>> asymptotics.main(potential='polynomial',
                ...     coefficients=[1, 2, 3],
                ...     varepsilon_list=[10, 25, 100])

    Example:
        Export .csv files for external use:

            >>> from ufjc.examples import asymptotics
            >>> asymptotics.main(potential='harmonic',
            ...     varepsilon_list=[10, 25, 100, 1000], csv=1)
            >>> asymptotics.main(potential='morse',
            ...     varepsilon_list=[10, 25, 100, 1000], csv=1)

    """

    # Import external modules
    import numpy as np
    import matplotlib.pyplot as plt

    # Import internal modules
    from ufjc import uFJC

    # Option to show plots or export .csv files
    export_csv_not_plot = kwargs.get('csv', 0) == 1
    if export_csv_not_plot is False:
        plt.figure()

    # Repeat for all values of varepsilon
    for varepsilon in kwargs.get('varepsilon_list'):

        # Function to produce a file name for .csv output
        def filename(model, approach):
            return './fig-' + str(model.potential) + '-' + \
                str(approach) + '-varepsilon-' + str(varepsilon) + '.csv'

        # Create the single-chain model
        model = uFJC(varepsilon=varepsilon, **kwargs)

        # If harmonic potentials, use the exact solution
        if model.potential == 'harmonic':
            eta_scale = model.varepsilon
            eta = np.linspace(1e-2, 0.3*model.varepsilon,
                              int(kwargs.get('num_points_eta', 100)))
            if export_csv_not_plot is True:
                data = np.vstack((model.gamma(eta, approach='exact'),
                                  eta/eta_scale)).T
                np.savetxt(filename(model, 'exact'), data, delimiter=",")
            else:
                plt.plot(model.gamma(eta, approach='exact'), eta/eta_scale,
                         label=r'$\varepsilon=$' + str(varepsilon))

        # Otherwise use a numerical calculation
        else:
            if model.potential == 'polynomial':
                eta_scale = model.varepsilon
            else:
                eta_scale = model.eta_max
            eta = np.linspace(1e-2, 0.99*eta_scale,
                              int(kwargs.get('num_points_eta', 100)))
            if export_csv_not_plot is True:
                data = np.vstack((model.gamma(eta, approach='quadrature'),
                                  eta/eta_scale)).T
                np.savetxt(filename(model, 'quad'), data, delimiter=",")
            else:
                plt.plot(model.gamma(eta, approach='quadrature'),
                         eta/eta_scale,
                         label=r'$\varepsilon=$' + str(varepsilon))

        # The asymptotic approaches
        if export_csv_not_plot is True:
            data = np.vstack((model.gamma(eta, approach='asymptotic'),
                              eta/eta_scale)).T
            np.savetxt(filename(model, 'asymptotic'), data, delimiter=",")
        else:
            plt.plot(model.gamma(eta, approach='asymptotic'),
                     eta/eta_scale, 'k:')

        # The reduced asymptotic results
        if export_csv_not_plot is True:
            data = np.vstack((model.gamma(eta, approach='reduced'),
                              eta/eta_scale)).T
            np.savetxt(filename(model, 'reduced'), data, delimiter=",")
        else:
            plt.plot(model.gamma(eta, approach='reduced'),
                     eta/eta_scale, 'k--')

    # Show the results if plotting
    if export_csv_not_plot is False:

        # Create linestyle (approach) legend
        if model.potential == 'harmonic':
            plt.plot(np.nan, np.nan, 'k', label='exact')
        else:
            plt.plot(np.nan, np.nan, 'k', label='quadrature')
        plt.plot(np.nan, np.nan, 'k:', label='asymptotic')
        plt.plot(np.nan, np.nan, 'k--', label='reduced')

        # Plot labels and legend
        plt.legend()
        plt.xlabel(r'$\gamma$')
        plt.ylabel(r'$\eta/\eta_\mathrm{max}$')
        plt.title(model.potential + '-FJC isotensional ' + r'$\gamma(\eta)$')
        plt.show()


if __name__ == '__main__':  # pragma: no cover

    # Import external modules
    import argparse

    # Parse args
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--potential', type=str, default='harmonic')
    parser.add_argument('-l', '--varepsilon_list', nargs='+', type=float,
                        required=True)

    # Parse arbitrary float args
    for arg in parser.parse_known_args()[1]:
        if arg.startswith('--'):
            parser.add_argument(arg.split('=')[0], type=float)

    # Execute main function with passed args as kwargs
    main(**vars(parser.parse_args()))



"""A module for handling potentials.

This module contains several different classes representing potentials,
each having methods to compute relevant nondimensional quantities as
functions of nondimensional force or stretch.
This module also contains the parent class ``Potential`` that is used to
assign a potential to a given model using keyword arguments.

Examples:
    Create a Lennard-Jones potential model with
    a nondimensional potential energy scale of 8 and evaluate
    the nondimensional potential energy at a stretch of 1.23:

        >>> from ufjc.potential import LennardJonesPotential
        >>> model = LennardJonesPotential(varepsilon=8)
        >>> model.beta_u(1.23)
        4.046654314368616

    Do the same with the Lenard-Jones-FENE potential:

        >>> from ufjc.potential import LJFENEPotential
        >>> model = LJFENEPotential(varepsilon=(8, 8))
        >>> model.beta_u(1.23)
        8.510502022381505

    Create a single-link model in one dimension, instantiate it with
    the Morse potential, and compute the incremental link stretch under
    a nondimensional force of 8:

        >>> from ufjc.potential import Potential
        >>> class Link1D(Potential):
        ...     def __init__(self, **kwargs):
        ...         Potential.__init__(self, **kwargs)
        >>> _ = Link1D(potential='morse').delta_lambda(8)
        >>> Link1D(potential='lj-fene').eta_link(1)
        184.0

"""

# Import external modules
import numpy as np
from scipy.special import lambertw


class HarmonicPotential(object):
    r"""The harmonic potential.

    Attributes:
        varepsilon (float): The nondimensional energy scale.
        kappa (float): The nondimensional stiffness :math:`\kappa=\varepsilon`.
        c (float): The correction parameter :math:`c=1`.

    """
    def __init__(self, **kwargs):
        """Initializes the ``HarmonicPotential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.
                Can be used to specify ``varepsilon`` (default 88).

        """
        self.varepsilon = kwargs.get('varepsilon', 88)
        self.kappa = self.varepsilon
        self.c = 1

    def phi(self, lambda_):
        r"""The scaled nondimensional potential energy function,

            .. math::
                \phi(\lambda) = \frac{1}{2}(\lambda-1)^2
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The scaled nondimensional potential energy(s).

        """
        return 0.5*(lambda_ - 1)**2

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy function,

            .. math::
                \beta u(\lambda) = \varepsilon\phi(\lambda)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional potential energy(s).

        """
        return self.varepsilon*self.phi(lambda_)

    def eta_link(self, lambda_):
        r"""The nondimensional force as a function of stretch,

            .. math::
                \eta(\lambda) = \varepsilon(\lambda - 1)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional force(s).

        Example:
            Compute the nondimensional force at a sample stretch:

                >>> from ufjc.potential import HarmonicPotential
                >>> HarmonicPotential().eta_link(1.8)
                70.4

        """
        return self.varepsilon*(lambda_ - 1)

    def delta_lambda(self, eta):
        r"""The incremental stretch as a function of nondimensional force,

            .. math::
                \Delta\lambda(\eta) = \frac{\eta}{\varepsilon}
                .

        Args:
            eta (array_like): The nondimensional force(s).

        Returns:
            numpy.ndarray: The incremental stretch(s).

        """
        return eta/self.varepsilon


class LogSquaredPotential(object):
    r"""The log-squared potential :cite:`mao2017rupture`.

    Attributes:
        varepsilon (float): The nondimensional energy scale.
        kappa (float): The nondimensional stiffness :math:`\kappa=\varepsilon`.
        c (float): The correction parameter :math:`c=2/5`.
        eta_max (float): The maximum nondimensional force
            :math:`\eta_\mathrm{max} = e^{-1}\varepsilon`.
        lambda_max (float): The stretch at the maximum nondimensional force,
            :math:`\lambda_\mathrm{max} = e^{1}`.

    """
    def __init__(self, **kwargs):
        """Initializes the ``LogSquaredPotential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.
                Can be used to specify ``varepsilon`` (default 88).

        """
        self.varepsilon = kwargs.get('varepsilon', 88)
        self.kappa = self.varepsilon
        self.c = 2/5
        self.eta_max = self.varepsilon/np.exp(1)
        self.lambda_max = np.exp(1)

    def phi(self, lambda_):
        r"""The scaled nondimensional potential energy function,

            .. math::
                \phi(\lambda) = \frac{1}{2}\big[\ln(\lambda)\big]^2
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The scaled nondimensional potential energy(s).

        """
        return 0.5*np.log(lambda_)**2

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy function,

            .. math::
                \beta u(\lambda) = \varepsilon\phi(\lambda)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional potential energy(s).

        """
        return self.varepsilon*self.phi(lambda_)

    def eta_link(self, lambda_):
        r"""The nondimensional force as a function of stretch,

            .. math::
                \eta(\lambda) = \varepsilon\,\frac{\ln(\lambda)}{\lambda}
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional force(s).

        Example:
            Compute the nondimensional force at a sample stretch:

                >>> from ufjc.potential import LogSquaredPotential
                >>> LogSquaredPotential().eta_link(1.8)
                28.736236950770266

        """
        return self.varepsilon*np.log(lambda_)/lambda_

    def delta_lambda(self, eta):
        r"""The incremental stretch as a function of nondimensional force,

            .. math::
                \Delta\lambda(\eta) = e^{-\mathcal{W}(-\eta/\varepsilon)}
                ,\qquad \eta\in[0,\eta_\mathrm{max}]
                .

        Args:
            eta (array_like): The nondimensional force(s).

        Returns:
            numpy.ndarray: The incremental stretch(s).

        """
        return (np.exp(-lambertw(-eta/self.varepsilon)) - 1).real


class MorsePotential(object):
    r"""The Morse potential :cite:`morse1929diatomic`.

    Attributes:
        varepsilon (float): The nondimensional energy scale.
        alpha (float): The Morse parameter.
        kappa (float): The nondimensional stiffness
            :math:`\kappa=2\varepsilon\alpha^2`.
        c (float): The correction parameter
            :math:`c=1/(1+3\alpha/2)`.
        eta_max (float): The maximum nondimensional force
            :math:`\eta_\mathrm{max} = \sqrt{\kappa\varepsilon/8}`.
        lambda_max (float): The stretch at the maximum nondimensional force,
            :math:`\lambda_\mathrm{max} = 1+\ln(2)/\alpha`.

    """
    def __init__(self, **kwargs):
        """Initializes the ``MorsePotential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.
                Can be used to specify ``varepsilon`` (default 88)
                and ``alpha`` (default 1).

        """
        self.varepsilon = kwargs.get('varepsilon', 88)
        self.alpha = kwargs.get('alpha', 1)
        self.kappa = 2*self.varepsilon*self.alpha**2
        self.c = 1/(1 + 3/2*self.alpha)
        self.eta_max = np.sqrt(self.kappa*self.varepsilon/8)
        self.lambda_max = 1 + np.log(2)/self.alpha

    def phi(self, lambda_):
        r"""The scaled nondimensional potential energy function,

            .. math::
                \phi(\lambda) = \left[1
                    - e^{-\alpha(\lambda - 1)}\right]^2
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The scaled nondimensional potential energy(s).

        """
        return (1 - np.exp(-self.alpha*(lambda_ - 1)))**2

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy function,

            .. math::
                \beta u(\lambda) = \varepsilon\phi(\lambda)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional potential energy(s).

        """
        return self.varepsilon*self.phi(lambda_)

    def eta_link(self, lambda_):
        r"""The nondimensional force as a function of stretch,

            .. math::
                \eta(\lambda) = 2\alpha\varepsilon e^{-\alpha(\lambda - 1)}
                    \left[1 - e^{-\alpha(\lambda - 1)}\right]^2
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional force(s).

        Example:
            Compute the nondimensional force at a sample stretch:

                >>> from ufjc.potential import MorsePotential
                >>> _ = MorsePotential().eta_link(1.23)

        """
        return 2*self.alpha*self.varepsilon * \
            np.exp(-self.alpha*(lambda_ - 1)) * \
            (1 - np.exp(-self.alpha*(lambda_ - 1)))

    def delta_lambda(self, eta):
        r"""The incremental stretch as a function of nondimensional force,

            .. math::
                \Delta\lambda(\eta) =
                    \ln\left(\frac{2}{1 + \sqrt{1 -
                        \eta/\eta_\mathrm{max}}}\right)^{1/\alpha}
                ,\qquad \eta\in[0,\eta_\mathrm{max}]
                .

        Args:
            eta (array_like): The nondimensional force(s).

        Returns:
            numpy.ndarray: The incremental stretch(s).

        """
        return np.log(
            2/(1 + np.sqrt(1 - eta/self.eta_max))
        )/self.alpha


class LennardJonesPotential(object):
    r"""The Lennard-Jones potential :cite:`jones1924determinationii`.

    Attributes:
        varepsilon (float): The nondimensional energy scale.
        kappa (float): The nondimensional stiffness
            :math:`\kappa=72\varepsilon`.
        c (float): The correction parameter
            :math:`c=2/23`.
        eta_max (float): The maximum nondimensional force
            :math:`\eta_\mathrm{max} = \eta(\lambda_\mathrm{max})`.
        lambda_max (float): The stretch at the maximum nondimensional force,
            :math:`\lambda_\mathrm{max} = (13/7)^{1/6}`.

    """
    def __init__(self, **kwargs):
        """Initializes the ``LennardJonesPotential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.
                Can be used to specify ``varepsilon`` (default 88).

        """
        self.varepsilon = kwargs.get('varepsilon', 88)
        self.kappa = 72*self.varepsilon
        self.c = 2/23
        self.lambda_max = (13/7)**(1/6)
        self.eta_max = self.eta_link(self.lambda_max)

    def phi(self, lambda_):
        r"""The scaled nondimensional potential energy function,

            .. math::
                \phi(\lambda) =
                    \frac{1}{\lambda^{12}} - \frac{2}{\lambda^6} + 1
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The scaled nondimensional potential energy(s).

        """
        return 1/lambda_**12 - 2/lambda_**6 + 1

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy function,

            .. math::
                \beta u(\lambda) = \varepsilon\phi(\lambda)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional potential energy(s).

        """
        return self.varepsilon*self.phi(lambda_)

    def eta_link(self, lambda_):
        r"""The nondimensional force as a function of stretch,

            .. math::
                \eta(\lambda) = 12\varepsilon\left(
                    \frac{1}{\lambda^7} - \frac{1}{\lambda^{13}}\right)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional force(s).

        """
        return self.varepsilon*(12/lambda_**7 - 12/lambda_**13)


class LJFENEPotential(object):
    r"""The Lennard-Jones-FENE potential :cite:`kremer1990dynamics`.

    Attributes:
        varepsilon (float): The nondimensional energy scale.
        kappa (float): The nondimensional stiffness.
        c (float): The correction parameter.
        eta_max (float): The maximum nondimensional force
            :math:`\eta_\mathrm{max} = \eta(\lambda_\mathrm{max})`.
        lambda_max (float): The stretch at the maximum nondimensional force.

    """
    def __init__(self, **kwargs):
        """Initializes the ``LJFENEPotential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.
                Can be used to specify ``varepsilon`` (default (88, 230))
                and ``lambda_max`` (default 1.5).

        """
        self.varepsilon_1, self.varepsilon_2 = \
            kwargs.get('varepsilon', (88, 230))
        self.varepsilon = self.varepsilon_2
        self.lambda_max = kwargs.get('lambda_max', 1.5)
        self.kappa = 72*self.varepsilon_1 + self.varepsilon_2 * \
            (self.lambda_max**2 + 1)/(self.lambda_max**2 - 1)**2
        self.c = 1/(1 - (
                -1512 + (6*self.lambda_max**2 + 2)/(self.lambda_max**2 - 1)**3
            )/(2*self.kappa/self.varepsilon_2)
        )
        self.eta_max = self.eta_link(self.lambda_max)

    def phi(self, lambda_):
        r"""The scaled nondimensional potential energy function,

            .. math::
                \phi(\lambda) =
                \frac{\varepsilon_1}{\varepsilon_2}\left(
                    \frac{1}{\lambda^{12}} - \frac{2}{\lambda^6} + 1
                \right) - \frac{1}{2}\,\ln\left[
                    1 - \left(\frac{\lambda}{\lambda_\mathrm{max}}\right)^2
                \right]
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The scaled nondimensional potential energy(s).

        """
        lambda_fene = (lambda_ < self.lambda_max)*lambda_
        return self.varepsilon_1/self.varepsilon_2*(
            1/lambda_**12 - 2/lambda_**6 + 1
        ) - 0.5*np.log(
            1 - (lambda_fene/self.lambda_max)**2
        )*(lambda_ < self.lambda_max)

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy function,

            .. math::
                \beta u(\lambda) =
                \varepsilon\phi(\lambda) =
                \varepsilon_1\left(
                    \frac{1}{\lambda^{12}} - \frac{2}{\lambda^6} + 1
                \right) - \frac{\varepsilon_2}{2}\,\ln\left[
                    1 - \left(\frac{\lambda}{\lambda_\mathrm{max}}\right)^2
                \right]
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional potential energy(s).

        """
        return self.varepsilon*self.phi(lambda_)

    def eta_link(self, lambda_):
        r"""The nondimensional force as a function of stretch,

            .. math::
                \eta(\lambda) = 12\varepsilon_1\left(
                    \frac{1}{\lambda^7} - \frac{1}{\lambda^{13}}
                \right) + \frac{\varepsilon_2\lambda}
                {\lambda_\mathrm{max}^2 - \lambda^2}
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional force(s).

        """
        lambda_fene = (lambda_ < self.lambda_max)*lambda_
        return self.varepsilon_1*(12/lambda_**7 - 12/lambda_**13) + \
            self.varepsilon_2*lambda_fene/(
                self.lambda_max**2 - lambda_fene**2
        )*(lambda_ < self.lambda_max)


class MiePotential(object):
    r"""The Mie potential :cite:`mie1903kinetischen`.

    Attributes:
        varepsilon (float): The nondimensional energy scale.
        n (float): The repulsive exponent.
        m (float): The attractive exponent.
        kappa (float): The nondimensional stiffness
            :math:`\kappa=nm\varepsilon`.
        c (float): The correction parameter
            :math:`c=\frac{4m(m+1)-2n(n+1)}{2m(m^2+5m+4)-n(n^2+5n+4)}`.
        eta_max (float): The maximum nondimensional force
            :math:`\eta_\mathrm{max} = \eta(\lambda_\mathrm{max})`.
        lambda_max (float): The stretch at the maximum nondimensional force,
            :math:`\lambda_\mathrm{max}=\left[\frac{n+1}{m+1}\right]^{1/(n-m)}`
            .

    """
    def __init__(self, **kwargs):
        """Initializes the ``MiePotential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.
                Can be used to specify ``varepsilon`` (default 88)
                ``n`` (default 12) and ``m`` (default 6).

        """
        self.varepsilon = kwargs.get('varepsilon', 88)
        self.m = kwargs.get('m', 6)
        self.n = kwargs.get('n', 12)
        self.kappa = self.n*self.m*self.varepsilon
        self.c = (
            4*self.m*(self.m + 1) - 2*self.n*(self.n + 1)
        )/(
            2*self.m*(self.m**2 + 5*self.m + 4) -
            self.n*(self.n**2 + 5*self.n + 4)
        )
        self.lambda_max = (
            (self.n + 1)/(self.m + 1)
        )**(
            1/(self.n - self.m)
        )
        self.eta_max = self.eta_link(self.lambda_max)

    def phi(self, lambda_):
        r"""The scaled nondimensional potential energy function,

            .. math::
                \phi(\lambda) = \frac{1}{(n-m)}
                    \left(\frac{m}{\lambda^n} - \frac{n}{\lambda^m}\right)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The scaled nondimensional potential energy(s).

        """
        return (
            self.m/lambda_**self.n - self.n/lambda_**self.m
        )/(
            self.n - self.m
        )

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy function,

            .. math::
                \beta u(\lambda) = \varepsilon\phi(\lambda)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional potential energy(s).

        """
        return self.varepsilon*self.phi(lambda_)

    def eta_link(self, lambda_):
        r"""The nondimensional force as a function of stretch,

            .. math::
                \eta(\lambda) = \varepsilon\, \frac{nm}{(n-m)}
                    \left(\frac{1}{\lambda^{m+1}}
                        - \frac{1}{\lambda^{n+1}}\right)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional force(s).

        """
        return self.varepsilon*self.n*self.m/(self.n - self.m)*(
            1/lambda_**(self.m + 1) - 1/lambda_**(self.n + 1)
        )


class PolynomialPotential(object):
    r"""A polynomial potential.

    Attributes:
        varepsilon (float): The nondimensional energy scale.
        coefficients (array_like): The coefficients :math:`a_k`.
        kappa (float): The nondimensional stiffness
            :math:`\kappa=a_1\varepsilon`.
        c (float): The correction parameter
            :math:`c=(1 - \frac{a_2}{2a_1})^{-1}`.

    """
    def __init__(self, **kwargs):
        """Initializes the ``PolynomialPotential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.
                Can be used to specify ``varepsilon`` (default 88)
                and ``coefficients`` (default 1).

        """
        self.varepsilon = kwargs.get('varepsilon', 88)
        coef = np.array(kwargs.get('coefficients', [1, 0]))
        self.eta_c = np.append(np.array([0]), coef)
        self.phi_c = np.append([0], self.eta_c)*np.append(
            [0, 0],
            [1/n for n in range(2, len(coef) + 2)]
        )
        self.kappa = self.varepsilon*self.eta_c[1]
        if len(self.phi_c) > 3:
            self.c = 1/(1 - self.phi_c[3]/self.phi_c[2]/2)

    def phi(self, lambda_):
        r"""The scaled nondimensional potential energy function,

            .. math::
                \phi(\lambda) = \sum_{k=2} \frac{a_{k-1}}{k}\, \lambda^k
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The scaled nondimensional potential energy(s).

        """
        return np.polynomial.Polynomial(self.phi_c)(lambda_ - 1)

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy function,

            .. math::
                \beta u(\lambda) = \varepsilon\phi(\lambda)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional potential energy(s).

        """
        return self.varepsilon*self.phi(lambda_)

    def eta_link(self, lambda_):
        r"""The nondimensional force as a function of stretch,

            .. math::
                \eta(\lambda) = \varepsilon\sum_{k=1} a_k\, \lambda^k
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional force(s).

        """
        return self.varepsilon * \
            np.polynomial.Polynomial(self.eta_c)(lambda_ - 1)


class CustomPotential(object):
    r"""A custom user-defined potential.

    Attributes:
        potential (str): The potential name.
        varepsilon (float): The nondimensional energy scale.
        phi (function): The scaled nondimensional energy function.
        eta_link (function): The nondimensional force as a function of stretch.
        delta_lambda (function): The incremental stretch as a function of the
            nondimensional force (optional).
        kappa (float): The nondimensional stiffness.
        c (float): The correction parameter.

    """
    def __init__(self, **kwargs):
        """Initializes the ``CustomPotential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.

        """
        self.potential = kwargs.get('potential', 'custom')
        self.varepsilon = kwargs.get('varepsilon')
        self.phi = kwargs.get('phi')
        self.eta_link = kwargs.get('eta_link')
        self.kappa = kwargs.get('kappa')
        self.c = kwargs.get('c')

        # No defaults
        if 'delta_lambda' in kwargs:
            self.delta_lambda = kwargs.get('delta_lambda')
        if 'lambda_max' in kwargs:
            self.lambda_max = kwargs.get('lambda_max')
        if 'eta_max' in kwargs:
            self.eta_max = kwargs.get('eta_max')

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy function,

            .. math::
                \beta u(\lambda) = \varepsilon\phi(\lambda)
                .

        Args:
            lambda_ (array_like): The stretch(s).

        Returns:
            numpy.ndarray: The nondimensional potential energy(s).

        """
        return self.varepsilon*self.phi(lambda_)


class Potential(object):
    r"""A class to assign a potential to a given model through inheritance.

    Attributes:
        potential (str): The potential type.
        pot (object): The potential model instance.

    """
    def __init__(self, **kwargs):
        """Initializes the ``Potential`` class.

        Args:
            **kwargs: Arbitrary keyword arguments.
                Used to specify the potential and then passed to
                an instantiation of that potential.

        Note:
            An improperly-specified potential will default to harmonic:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='blah')
                Potential "blah" invalid, defaulting to "harmonic" potential.

        """
        self.potential = kwargs.get('potential', 'harmonic')
        if self.potential == 'harmonic':
            self.pot = HarmonicPotential(**kwargs)
        elif self.potential == 'log-squared':
            self.pot = LogSquaredPotential(**kwargs)
        elif self.potential == 'morse':
            self.pot = MorsePotential(**kwargs)
        elif self.potential == 'lennard-jones':
            self.pot = LennardJonesPotential(**kwargs)
        elif self.potential == 'lj-fene':
            self.pot = LJFENEPotential(**kwargs)
        elif self.potential == 'mie':
            self.pot = MiePotential(**kwargs)
        elif self.potential == 'polynomial':
            self.pot = PolynomialPotential(**kwargs)
        elif self.potential == 'custom':
            self.pot = CustomPotential(**kwargs)
        else:
            print('Potential "' + self.potential +
                  '" invalid, defaulting to "harmonic" potential.')
            self.pot = HarmonicPotential(**kwargs)

    def __getattr__(self, attr):
        """Inherit attributes and methods from the chosen potential.

            Note:
                When accessing an attribute of the model to which the potential
                is assigned, but no attribute is found, and ``AttributeError``
                will be raised with respect to the ``pot`` instance, not the
                overall model instance. This is sub-optimal but typically
                acceptable, but still should be repaired in the future.

        """
        return getattr(self.pot, attr)



"""A module for the uFJC single-chain model in the isometric ensemble.

This module consist of the class ``uFJCIsometric`` which contains
methods for computing single-chain quantities
in the isometric (constant end-to-end vector) thermodynamic ensemble.

Example:
    Import and instantiate the class:

        >>> from ufjc.isometric import uFJCIsometric
        >>> class_instance = uFJCIsometric()

"""

# Import internal modules
from .monte_carlo import MHMCMC
from .isotensional import uFJCIsotensional

# Import external modules
import numpy as np
import numpy.linalg as la


class uFJCIsometric(uFJCIsotensional):
    """The uFJC single-chain model class for the isometric ensemble.

    This class contains methods for computing single-chain quantities
    in the isometric (constant end-to-end vector) thermodynamic ensemble.
    It inherits all attributes and methods from the ``uFJCIsotensional``
    class, which inherits all attributes and methods from the
    ``BasicUtility`` class.

    """
    def __init__(self):
        """Initializes the ``uFJCIsometric`` class.

        Initialize and inherit all attributes and methods
        from a ``uFJCIsotensional`` class instance.

        """
        uFJCIsotensional.__init__(self)

    def eta_isometric(self, gamma, **kwargs):
        r"""Main function for the isometric :math:`\eta(\gamma)`.

        This is the main function utilized to compute the isometric
        nondimensional single-chain mechanical response.
        Keyword arguments specify and are passed onto the method.

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to the chosen method.

        Returns:
            numpy.ndarray: The nondimensional force(s).

        Example:
            Compute the nondimensional force for an eight-link Morse-FJC at a
            nondimensional end-to-end length of 0.8 in the isometric ensemble,
            using the Legendre transformation method from the isotensional
            ensemble, and using the reduced asymptotic approach to compute
            quantities in the isotensional ensemble:

                >>> from ufjc import uFJC
                >>> model = uFJC(N_b=8, potential='morse')
                >>> model.eta_isometric([0, 0.8], \
                ...     method='legendre', approach='reduced')
                array([0.        , 4.41715473])

        Warning:
            Only the Legendre transformation method is currently unavailable:

                >>> from ufjc import uFJC
                >>> uFJC().eta_isometric(0.8, method='exact')
                array([nan])

        """
        gamma = self.np_array(gamma)
        method = kwargs.get('method', 'legendre')
        if method == 'legendre':
            return self.eta_isometric_legendre(gamma, **kwargs)
        else:
            return np.nan*gamma

    def eta_isometric_legendre(self, gamma, **kwargs):
        r"""The Legendre transformation method of approximating
        the isometric :math:`\eta(\gamma)`.

        This function uses the Legendre transformation method to obtain an
        approximate isometric nondimensional single-chain mechanical response.
        The result is to simply use the isotensional :math:`\eta(\gamma)`,
        and this approximation is asymptotically valid for :math:`N_b\gg 1`
        and appreciable loads :cite:`buche2020statistical`.

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to ``_eta_isotensional``.

        Returns:
            numpy.ndarray: The nondimensional force(s).

        Example:
            Compute the nondimensional force at a large nondimensional
            end-to-end length using the Legendre transformation method:

                >>> from ufjc import uFJC
                >>> model = uFJC()
                >>> model.eta_isometric_legendre(1.3)
                array([28.71102552])

        """
        return self.eta_isotensional(gamma, **kwargs)

    def gamma_isometric(self, eta, **kwargs):
        r"""Main function for the isometric :math:`\gamma(\eta)`.

        This function obtains the isometric nondimensional
        single-chain mechanical response :math:`\gamma(\eta)`
        by inverting the isometric :math:`\eta(\gamma)`.

        Args:
            eta (array_like): the nondimensional force(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to ``_eta_isometric``.

        Returns:
            numpy.ndarray: The nondimensional end-to-end length(s).

        Example:
            Check that :math:`\gamma[\eta(\gamma)] = \gamma\,`:

                >>> import numpy as np
                >>> from ufjc import uFJC
                >>> model = uFJC()
                >>> def check_eta(gamma):
                ...     eta_fun = lambda gamma: model.eta_isometric(gamma)
                ...     gamma_fun = lambda eta: model.gamma_isometric(eta)
                ...     return np.isclose(gamma_fun(eta_fun(gamma))[0], gamma)
                >>> check_eta(np.random.rand())
                True

        """
        def eta_fun(gamma):
            return self.eta_isometric(gamma, **kwargs)
        return self.inv_fun_1D(eta, eta_fun)

    def vartheta_isometric(self, gamma, **kwargs):
        r"""Main function for the isometric :math:`\vartheta(\gamma)`.

        This is the main function utilized to compute the nondimensional
        Helmholtz free energy per link, an isometric quantity.
        Keyword arguments specify and are passed onto the method.

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to the chosen method.

        Returns:
            numpy.ndarray: The nondimensional Helmholtz free energy per link.

        Example:
            Compute the nondimensional Helmholtz free energy per link
            for an eight-link Morse-FJC at a
            nondimensional end-to-end length of 0.8 in the isometric ensemble,
            using the Legendre transformation method from the isotensional
            ensemble, and using the reduced asymptotic approach to compute
            quantities in the isotensional ensemble:

                >>> from ufjc import uFJC
                >>> model = uFJC(N_b=8, potential='morse')
                >>> model.vartheta_isometric(0.8, \
                ...     method='legendre', approach='reduced')
                array([1.23847534])

        Warning:
            The exact method is currently unavailable:

                >>> from ufjc import uFJC
                >>> uFJC().vartheta_isometric(0.8, method='exact')
                nan

        """
        method = kwargs.get('method', 'legendre')
        if method == 'exact':
            return np.nan*gamma
        elif method == 'legendre':
            return self.vartheta_isometric_legendre(gamma, **kwargs)

    def vartheta_isometric_legendre(self, gamma, **kwargs):
        r"""The Legendre transformation method of approximating
        the isometric :math:`\vartheta(\gamma)`.

        This function uses the Legendre transformation method to obtain an
        approximate isometric Helmholtz free energy per link.
        The result is independent of the number of links :math:`N_b`, and
        this approximation is asymptotically valid for :math:`N_b\gg 1`
        and appreciable loads :cite:`buche2021chain`.
        For example, using the reduced asymptotic approach, this is

        .. math::
            \vartheta(\gamma) \sim
            \ln\left\{\frac{
            \eta\exp[\eta\mathcal{L}(\eta)]}{\sinh(\eta)}\right\}
            + \beta u[\lambda(\eta)]
            ,

        valid when :math:`\varepsilon\gg 1` and :math:`N_b\gg 1` are
        simultaneously true.
        Note that :math:`\eta=\eta(\gamma)` is implied, and obtained
        through inverting the isotensional :math:`\gamma(\eta)`.

        Args:
            gamma (array_like): The nondimensional end-to-end length(s).
            **kwargs: Arbitrary keyword arguments.
                Passed to the chosen method.

        Returns:
            numpy.ndarray: The nondimensional Helmholtz free energy per link.

        Example:
            Approximate the nondimensional Helmholtz free energy per link
            using the Legendre method and both asymptotic approaches:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='log-squared', varepsilon=23)
                >>> model.vartheta_isometric_legendre(1.1)
                array([1.90431381])
                >>> model.vartheta_isometric_legendre(1.1, approach='reduced')
                array([2.09238198])

        Warning:
            Only the asymptotic approaches are currently unavailable:

                >>> from ufjc import uFJC
                >>> model = uFJC(potential='log-squared', varepsilon=23)
                >>> model.vartheta_isometric_legendre(1.1, approach='exact')
                nan

        """
        # Invert gamma=gamma(eta) for the corresponding eta
        eta = self.eta_isotensional(gamma, **kwargs)

        # Avoid overflow, important for integrating P_eq
        eta[eta > self.maximum_exponent] = self.maximum_exponent

        # Find the corresponding bond stretch under direct eta
        lambda_ = 1 + self.delta_lambda(eta)

        # Need to finish this portion
        approach = kwargs.get('approach', 'asymptotic')
        if approach == 'asymptotic':
            Ln = self.langevin(eta)
            coth = self.coth(eta)
            return eta*Ln + self.log_over_sinh(eta) + \
                self.varepsilon*(self.phi(lambda_) - self.phi(1)) + \
                eta**2/self.kappa*(
                    (1 - Ln*coth)/(self.c + eta/self.kappa*coth)
                ) - np.log(1 + eta*coth/self.kappa)
        elif approach == 'reduced':
            return eta*self.langevin(eta) + self.log_over_sinh(eta) + \
                self.varepsilon*(self.phi(lambda_) - self.phi(1))
        else:
            return np.nan*gamma

    def beta_U_config(self, config):
        r"""The nondimensional potential energy of a configuration.

        This function provides the nondimensional potential energy
        :math:`\beta U` given the configuration of the chain, i.e. the
        vector position of each atom/hinge relative to the first one.

        Args:
            config (numpy.ndarray): The configuration of the chain,
                a :math:`(N_b+1)`-by-3 numpy array.

        Returns:
            float: The nondimensional potential energy :math:`\beta U`.

        Example:
            Compute the potential energy of the uniformly-stretched
            default initial configuration:

                >>> from ufjc import uFJC
                >>> model = uFJC(N_b=8, potential='lennard-jones')
                >>> model.beta_U_config(1.1*model.init_config)
                133.5368021523727

        """
        beta_U = 0
        for j in range(1, len(config)):
            lambda_ = la.norm(config[j, :] - config[j - 1, :])
            beta_U += self.beta_u(lambda_)
        return beta_U



"""A module for testing examples within docstrings.

This module checks all other modules in the package
and ensures that the examples within each docstring
are executed properly and obtain the expected output.

"""

# Import external modules
import doctest
import unittest
from os import path
from glob import glob
from sys import platform


class TestDocstringExamples(unittest.TestCase):
    """Class to test examples within docstrings.

    """
    def test_docstring_python_examples(self, examples=True):
        """Function to test examples within docstrings.

       Args:
           examples (bool, optional, default=True): Whether to test examples.

        Returns:
            int: The total number of failures from all files.

        """
        tests_dir = path.dirname(__file__)
        files = glob(path.join(tests_dir, './*.py'))
        files += glob(path.join(tests_dir, '../*.py'))
        if examples:
            files += glob(path.join(tests_dir, '../examples/*.py'))
        else:  # pragma: no cover
            if platform == 'win32' or platform == 'win64':
                files.remove(glob(path.join(tests_dir, '..\\swfjc.py'))[0])
            else:
                files.remove(glob(path.join(tests_dir, '../swfjc.py'))[0])
        for file in files:
            self.assertFalse(doctest.testfile(file, module_relative=False)[0])


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import re
from os.path import join
from distutils.core import setup
from setuptools import find_packages


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content


def get_version():
    VERSIONFILE = join('ufjc', '__init__.py')
    with open(VERSIONFILE, 'rt') as f:
        lines = f.readlines()
    vgx = '^__version__ = \"[0-9+.0-9+.0-9+]*[a-zA-Z0-9]*\"'
    for line in lines:
        mo = re.search(vgx, line, re.M)
        if mo:
            return mo.group().split('"')[1]
    raise RuntimeError('Unable to find version in %s.' % (VERSIONFILE,))


setup(
    name='ufjc',
    version=get_version(),
    package_dir={'ufjc': 'ufjc'},
    packages=find_packages(),
    description='The Python package for the uFJC single-chain model.',
    long_description=read("README.rst"),
    author='Michael R. Buche, Scott J. Grutzik',
    author_email='mrbuche@sandia.gov, sjgrutz@sandia.gov',
    url='https://sandialabs.github.io/uFJC',
    license='BSD-3-Clause',
    keywords=['ufjc', 'polymers', 'statistical mechanics', 'thermodynamics'],
    install_requires=['numpy', 'pycodestyle', 'scipy'],
    extras_require={
      'docs': ['matplotlib', 'sphinx', 'sphinx-copybutton',
               'sphinx-rtd-theme', 'sphinxcontrib-bibtex'],
      'plotting': ['matplotlib'],
      'testing': ['matplotlib', 'pytest', 'pytest-cov'],
      'all': ['matplotlib', 'pytest', 'pytest-cov',
              'sphinx', 'sphinx-copybutton',
              'sphinx-rtd-theme', 'sphinxcontrib-bibtex']
    },
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    project_urls={
      'Anaconda': 'https://anaconda.org/mrbuche/ufjc',
      'Documentation': 'https://ufjc.readthedocs.io',
      'GitHub': 'https://github.com/sandialabs/ufjc',
      'Issues': 'https://github.com/sandialabs/ufjc/issues',
    },
)



"""Module-level functionality for testing installation.

Example:
    Test that the package was installed properly:

    ::

        python -m ufjc.tests

"""

from .test_doc import TestDocstringExamples  # pragma: no cover
inst = TestDocstringExamples()  # pragma: no cover
inst.test_docstring_python_examples(examples=False)  # pragma: no cover
